# Copyright (c) 2014 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading

from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import excutils
import requests
from requests import sessions

from networking_odl.common import utils

LOG = log.getLogger(__name__)
cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')


class OpenDaylightRestClient(object):
    @staticmethod
    def _check_opt(url):
        if not url:
            raise cfg.RequiredOptError('url', cfg.OptGroup('ml2_odl'))
        required_opts = ('url', 'username', 'password')
        for opt in required_opts:
            if not getattr(cfg.CONF.ml2_odl, opt):
                raise cfg.RequiredOptError(opt, cfg.OptGroup('ml2_odl'))

    @classmethod
    def create_client(cls, url=None):
        if cfg.CONF.ml2_odl.enable_lightweight_testing:
            LOG.debug("ODL lightweight testing is enabled, "
                      "returning a OpenDaylightLwtClient instance")

            # Have to import at here, otherwise we create a dependency loop
            # pylint: disable=import-outside-toplevel
            from networking_odl.common import lightweight_testing as lwt
            cls = lwt.OpenDaylightLwtClient

        url = url or cfg.CONF.ml2_odl.url
        cls._check_opt(url)
        return cls(
            url,
            cfg.CONF.ml2_odl.username,
            cfg.CONF.ml2_odl.password,
            cfg.CONF.ml2_odl.timeout)

    def __init__(self, url, username, password, timeout):
        super(OpenDaylightRestClient, self).__init__()
        self.url = url
        self.timeout = timeout
        self.session = sessions.Session()
        self.session.auth = (username, password)

    def get_resource(self, resource_type, resource_id):
        response = self.get(utils.make_url_object(resource_type) + '/' +
                            resource_id)
        if response.status_code == requests.codes.not_found:
            return None

        return self._check_response(response).json()

    def get(self, urlpath='', data=None):
        return self.request('get', urlpath, data)

    def request(self, method, urlpath='', data=None):
        headers = {'Content-Type': 'application/json'}
        url = '/'.join([self.url, urlpath])
        LOG.debug(
            "Sending METHOD (%(method)s) URL (%(url)s) JSON (%(data)s)",
            {'method': method, 'url': url, 'data': data})
        return self.session.request(
            method, url=url, headers=headers, data=data, timeout=self.timeout)

    def sendjson(self, method, urlpath, obj):
        """Send json to the OpenDaylight controller."""
        data = jsonutils.dumps(obj, indent=2) if obj else None
        try:
            return self._check_response(
                self.request(method, urlpath, data))
        except Exception:
            with excutils.save_and_reraise_exception():
                LOG.error("REST request ( %(method)s ) to "
                          "url ( %(urlpath)s ) is failed. "
                          "Request body : [%(body)s] service",
                          {'method': method,
                           'urlpath': urlpath,
                           'body': obj})
                return None

    def _check_response(self, response):
        try:
            response.raise_for_status()
        except requests.HTTPError as error:
            with excutils.save_and_reraise_exception():
                LOG.debug("Exception from ODL: %(e)s %(text)s",
                          {'e': error, 'text': response.text}, exc_info=1)
                return None
        else:
            LOG.debug("Got response:\n"
                      "(%(response)s)", {'response': response.text})
            return response


class OpenDaylightRestClientGlobal(object):
    """ODL Rest client as global variable

    The creation of OpenDaylightRestClient needs to be delayed until
    configuration values need to be configured at first.
    """
    def __init__(self):
        super(OpenDaylightRestClientGlobal, self).__init__()
        self._lock = threading.Lock()
        self._client = None

    def get_client(self):
        with self._lock:
            if self._client is None:
                self._client = OpenDaylightRestClient.create_client()
            return self._client



# Copyright (c) 2018 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.objects import router as l3_obj
from neutron_lib.callbacks import events
from neutron_lib.callbacks import resources
from neutron_lib.db import api as db_api
from oslo_config import fixture as config_fixture
from oslo_utils import uuidutils

from networking_odl.common import constants as odl_const
from networking_odl.db import db
from networking_odl.l3 import l3_flavor
from networking_odl.tests import base
from networking_odl.tests.unit import base_v2

_operation_map = {'del': odl_const.ODL_DELETE,
                  'update': odl_const.ODL_UPDATE,
                  'add': odl_const.ODL_CREATE}


class OpenDaylightL3FlavorTestCase(base_v2.OpenDaylightConfigBase):
    def setUp(self):
        self.useFixture(base.OpenDaylightJournalThreadFixture())
        self.cfg = self.useFixture(config_fixture.Config())
        self.cfg.config(service_plugins=['router'])
        super(OpenDaylightL3FlavorTestCase, self).setUp()
        self.flavor_driver = l3_flavor.ODLL3ServiceProvider(mock.MagicMock())

    def _get_mock_fip_kwargs(self):
        fipid = uuidutils.generate_uuid()
        fip_db = mock.Mock(floating_ip_address='192.168.1.2',
                           router_id=None, id=fipid,
                           floating_network_id=fipid)
        projectid = uuidutils.generate_uuid()
        floating_data = {'floatingip_id': str(fipid),
                         'router_id': None,
                         'context': self.db_context,
                         'floatingip_db': fip_db,
                         'floatingip': {'project_id': str(projectid),
                                        'floating_ip_address': '172.24.5.4',
                                        'port_id': None,
                                        'id': fipid,
                                        'router_id': None,
                                        'status': 'DOWN',
                                        'floating_network_id': str(fipid)
                                        }}
        return floating_data

    def _get_mock_router_kwargs(self):
        router_db = mock.Mock(gw_port_id=uuidutils.generate_uuid(),
                              id=uuidutils.generate_uuid())
        router = {odl_const.ODL_ROUTER:
                  {'name': 'router1',
                   'admin_state_up': True,
                   'tenant_id': uuidutils.generate_uuid(),
                   'id': router_db.id,
                   'external_gateway_info': {'network_id':
                                             uuidutils.generate_uuid()}},
                  'context': self.db_context,
                  "router_db": router_db}

        return router

    def _test_fip_operation(self, event, operation, fip, ops=True):
        method = getattr(self.flavor_driver,
                         '_floatingip_%s_%s' % (operation, event))
        state_0 = None
        if 'floatingip' in fip:
            state_0 = fip['floatingip']
        elif 'port' in fip:
            state_0 = fip['port']
        payload = events.DBEventPayload(
            context=self.context, resource_id=fip['floatingip_id'],
            states=(state_0, ),
            metadata={'floatingip_db': fip.get('floatingip_db')},
            desired_state=fip.get('floatingip_db'))
        method(odl_const.ODL_FLOATINGIP, event, mock.ANY, payload)
        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        if ops:
            if operation != odl_const.ODL_DELETE:
                self.assertEqual(fip['floatingip'], row.data)
            self.assertEqual(odl_const.ODL_FLOATINGIP, row.object_type)
            self.assertEqual(fip['floatingip_id'], row.object_uuid)
        else:
            self.assertIsNone(row)

    def _test_router_operation(self, event, operation, router, ops=True):
        method = getattr(self.flavor_driver,
                         '_router_%s_%s' % (operation, event))
        with db_api.CONTEXT_WRITER.using(self.db_context):
            if event == 'precommit':
                method(odl_const.ODL_ROUTER, mock.ANY, mock.ANY, **router)
            else:
                payload = events.DBEventPayload(
                    router.get('context'), states=(router.get('router_db'),),
                    request_body=router.get(resources.ROUTER),
                    resource_id=router.get(resources.ROUTER).get('id'))

                method(odl_const.ODL_ROUTER, mock.ANY, mock.ANY,
                       payload=payload)
        row = None
        rows = sorted(db.get_all_db_rows_by_state(self.db_context,
                                                  odl_const.PENDING),
                      key=lambda x: x.seqnum)
        if ops:
            row = rows[0]
            if operation in ['del', odl_const.ODL_DELETE]:
                self.assertEqual(router['router_id'], row.object_uuid)
            else:
                self.assertEqual(router['router'], row.data)
            self.assertEqual(_operation_map[operation], row.operation)
        else:
            self.assertIsNone(row)

    def test_router_add_association(self):
        with mock.patch.object(self.flavor_driver,
                               '_validate_l3_flavor',
                               return_value=True):
            router = self._get_mock_router_kwargs()
            # Driver Association payload is different and expects
            # router_id
            router['router_id'] = router['router']['id']
            self._test_router_operation("association", "add", router)

    def test_l3_operations_for_different_flavor(self):
        with mock.patch.object(self.flavor_driver,
                               '_validate_l3_flavor',
                               return_value=False):
            router = self._get_mock_router_kwargs()
            router['router_id'] = router['router']['id']
            self._test_router_operation("association", "add", router, False)
            self._test_router_operation("association", "del", router, False)

    def test_l3_router_update_precommit(self):
        with mock.patch.object(self.flavor_driver,
                               '_validate_l3_flavor',
                               return_value=True):
            router = self._get_mock_router_kwargs()
            router['router_id'] = router['router']['id']
            self._test_router_operation("precommit", "update", router)

    def test_router_del_association(self):
        with mock.patch.object(self.flavor_driver,
                               '_validate_l3_flavor',
                               return_value=True):
            router = self._get_mock_router_kwargs()
            router['router_id'] = router['router']['id']
            self._test_router_operation("association", "del", router)

    def test_fip_precommit_create(self):
        with mock.patch.object(self.flavor_driver,
                               '_validate_l3_flavor',
                               return_value=True):
            fip = self._get_mock_fip_kwargs()
            self._test_fip_operation("precommit", odl_const.ODL_CREATE, fip)

    def test_l3_fip_different_flavor(self):
        with mock.patch.object(self.flavor_driver,
                               '_validate_l3_flavor',
                               return_value=False):
            fip = self._get_mock_fip_kwargs()
            fip['old_floatingip'] = fip['floatingip']
            self._test_fip_operation("precommit",
                                     odl_const.ODL_CREATE, fip, False)
            self._test_fip_operation("precommit",
                                     odl_const.ODL_UPDATE, fip, False)

    def test_fip_precommit_delete(self):
        # As precommit delete gets port data
        fip = self._get_mock_fip_kwargs()
        port = {'port': {'id': uuidutils.generate_uuid()},
                'context': self.db_context,
                'floatingip_id': fip['floatingip_id']}
        with mock.patch.object(l3_obj.FloatingIP, 'get_objects',
                               return_value=[fip['floatingip_db']]):
            with mock.patch.object(self.flavor_driver,
                                   '_validate_l3_flavor',
                                   return_value=True):
                self._test_fip_operation("precommit",
                                         odl_const.ODL_DELETE, port)

    def test_fip_precommit_update(self):
        with mock.patch.object(self.flavor_driver,
                               '_validate_l3_flavor',
                               return_value=True):
            fip = self._get_mock_fip_kwargs()
            self._test_fip_operation("precommit", odl_const.ODL_UPDATE, fip)



# Copyright (c) 2017 NEC Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions

from neutron._i18n import _


class NetworkingODLException(exceptions.NeutronException):
    """Base Networking-ODL exception."""
    pass


class FullSyncError(NetworkingODLException):
    """Base exception for Full Sync"""
    pass


class UnsupportedResourceType(NetworkingODLException):
    """An exception for unsupported resource for full sync and recovery"""
    message = _("unsupported resource type: %(resource)s")


class PluginMethodNotFound(NetworkingODLException, AttributeError):
    """An exception indicating plugin method was not found.

       Specialization of AttributeError and NetworkingODLException indicating
       requested plugin method could not be found.

       :param method: Name of the method being accessed.
       :param plugin: Plugin name expected to have required method.
    """
    message = _("%(method)s not found in %(plugin)s")


class ResourceNotRegistered(FullSyncError):
    """An exception indicating resource is not registered for maintenance task.

       Specialization of FullSync error indicating resource is not registered
       for maintenance tasks full sync and recovery.

       :param resource_type: Resource type not registered for maintenance task.
    """
    message = _("%(resource_type)s resource is not registered for maintenance")



#    Copyright 2012 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import compiler
import importlib.util
import os.path
import sys


def is_localized(node):
    """Check message wrapped by _()"""
    if isinstance(node.parent, compiler.ast.CallFunc):
        if isinstance(node.parent.node, compiler.ast.Name):
            if node.parent.node.name == '_':
                return True
    return False


class ASTWalker(compiler.visitor.ASTVisitor):

    def default(self, node, *args):
        for child in node.getChildNodes():
            child.parent = node
        compiler.visitor.ASTVisitor.default(self, node, *args)


class Visitor(object):

    def __init__(self, filename, i18n_msg_predicates,
                 msg_format_checkers, debug):
        self.filename = filename
        self.debug = debug
        self.error = 0
        self.i18n_msg_predicates = i18n_msg_predicates
        self.msg_format_checkers = msg_format_checkers
        with open(filename) as f:
            self.lines = f.readlines()

    def visitConst(self, node):
        if not isinstance(node.value, str):
            return

        if is_localized(node):
            for (checker, msg) in self.msg_format_checkers:
                if checker(node):
                    print('%s:%d %s: %s Error: %s' %
                          (self.filename, node.lineno,
                           self.lines[node.lineno - 1][:-1],
                           checker.__name__, msg),
                           file=sys.stderr)
                    self.error = 1
                    return
            if debug:
                print('%s:%d %s: %s' %
                      (self.filename, node.lineno,
                       self.lines[node.lineno - 1][:-1],
                       "Pass"))
        else:
            for (predicate, action, msg) in self.i18n_msg_predicates:
                if predicate(node):
                    if action == 'skip':
                        if debug:
                            print('%s:%d %s: %s' %
                                  (self.filename, node.lineno,
                                  self.lines[node.lineno - 1][:-1],
                                  "Pass"))
                        return
                    elif action == 'error':
                        print('%s:%d %s: %s Error: %s' %
                              (self.filename, node.lineno,
                               self.lines[node.lineno - 1][:-1],
                               predicate.__name__, msg),
                               file=sys.stderr)
                        self.error = 1
                        return
                    elif action == 'warn':
                        print('%s:%d %s: %s' %
                              (self.filename, node.lineno,
                              self.lines[node.lineno - 1][:-1],
                              "Warn: %s" % msg))
                        return
                    print('Predicate with wrong action!', file=sys.stderr)


def is_file_in_black_list(black_list, f):
    for f in black_list:
        if os.path.abspath(input_file).startswith(
            os.path.abspath(f)):
            return True
    return False


def check_i18n(input_file, i18n_msg_predicates, msg_format_checkers, debug):
    input_mod = compiler.parseFile(input_file)
    v = compiler.visitor.walk(input_mod,
                              Visitor(input_file,
                                      i18n_msg_predicates,
                                      msg_format_checkers,
                                      debug),
                              ASTWalker())
    return v.error


def load_module(name, path):
    module_spec = importlib.util.spec_from_file_location(
        name, path
    )
    module = importlib.util.module_from_spec(module_spec)
    module_spec.loader.exec_module(module)
    return module


if __name__ == '__main__':
    input_path = sys.argv[1]
    cfg_path = sys.argv[2]
    try:
        cfg_mod = load_module('', cfg_path)
    except Exception:
        print("Load cfg module failed", file=sys.stderr)
        sys.exit(1)

    i18n_msg_predicates = cfg_mod.i18n_msg_predicates
    msg_format_checkers = cfg_mod.msg_format_checkers
    black_list = cfg_mod.file_black_list

    debug = False
    if len(sys.argv) > 3:
        if sys.argv[3] == '-d':
            debug = True

    if os.path.isfile(input_path):
        sys.exit(check_i18n(input_path,
                            i18n_msg_predicates,
                            msg_format_checkers,
                            debug))

    error = 0
    for dirpath, dirs, files in os.walk(input_path):
        for f in files:
            if not f.endswith('.py'):
                continue
            input_file = os.path.join(dirpath, f)
            if is_file_in_black_list(black_list, input_file):
                continue
            if check_i18n(input_file,
                          i18n_msg_predicates,
                          msg_format_checkers,
                          debug):
                error = 1
    sys.exit(error)



# Copyright (c) 2017 Ericsson India Global Service Pvt Ltd.
# All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib import context
from neutron_lib.plugins import directory
from neutron_lib.services.trunk import constants as t_consts
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron.services.trunk.drivers import base as trunk_base

from networking_odl.common import config as odl_conf
from networking_odl.common import constants as odl_const
from networking_odl.journal import full_sync
from networking_odl.journal import journal
from networking_odl.trunk import constants as odltrunk_const

LOG = logging.getLogger(__name__)

TRUNK_RESOURCES = {
    odl_const.ODL_TRUNK: odl_const.ODL_TRUNKS
}


@registry.has_registry_receivers
class OpenDaylightTrunkHandlerV2(object):
    def __init__(self):
        cfg.CONF.register_opts(odl_conf.odl_opts, "ml2_odl")
        self.journal = journal.OpenDaylightJournalThread()
        full_sync.register(resources.TRUNK, TRUNK_RESOURCES)
        LOG.info('initialized trunk driver for OpendayLight')

    @staticmethod
    def _record_in_journal(context, trunk_id, operation, data):
        journal.record(context, odl_const.ODL_TRUNK, trunk_id, operation, data)

    # TODO(vthapar) Revisit status updates once websockets are fully
    # implemented - https://review.opendev.org/#/c/421127/
    @log_helpers.log_method_call
    def trunk_create_precommit(self, resource, event,
                               trunk_plugin, payload=None):
        data = payload.latest_state.to_dict()
        data['status'] = t_consts.TRUNK_ACTIVE_STATUS
        self._record_in_journal(payload.context, payload.resource_id,
                                odl_const.ODL_CREATE, data)

    @log_helpers.log_method_call
    def trunk_update_precommit(self, resource, event,
                               trunk_plugin, payload=None):
        trunk_payload = payload.desired_state or payload.latest_state
        trunk_payload.update(status=t_consts.TRUNK_ACTIVE_STATUS)
        data = trunk_payload.to_dict()
        trunk_id = payload.resource_id

        self._record_in_journal(payload.context, trunk_id,
                                odl_const.ODL_UPDATE, data)

    @log_helpers.log_method_call
    def trunk_delete_precommit(self, resource, event,
                               trunk_plugin, payload=None):
        # fill in data with parent ids, will be used in parent validations
        trunk_dict = payload.states[0].to_dict()
        data = [subport['port_id'] for subport in trunk_dict['sub_ports']]
        data.append(trunk_dict['port_id'])
        self._record_in_journal(payload.context, payload.resource_id,
                                odl_const.ODL_DELETE, data)

    @log_helpers.log_method_call
    def trunk_create_postcommit(self, resource, event,
                                trunk_plugin, payload=None):
        payload.latest_state.update(status=t_consts.TRUNK_ACTIVE_STATUS)
        self.journal.set_sync_event()

    @log_helpers.log_method_call
    def trunk_update_postcommit(self, resource, event,
                                trunk_plugin, payload=None):
        payload.latest_state.update(status=t_consts.TRUNK_ACTIVE_STATUS)
        self.journal.set_sync_event()

    @log_helpers.log_method_call
    def trunk_delete_postcommit(self, resource, event,
                                trunk_plugin, payload=None):
        self.journal.set_sync_event()

    @log_helpers.log_method_call
    def trunk_subports_set_status(self, resource, event, trunk_plugin,
                                  payload=None):
        core_plugin = directory.get_plugin()
        admin_context = context.get_admin_context()

        if event == events.AFTER_DELETE:
            status = n_const.PORT_STATUS_DOWN
        else:
            parent_id = payload.latest_state.port_id
            parent_port = core_plugin._get_port(admin_context, parent_id)
            status = parent_port['status']

        for subport in payload.subports:
            self._set_subport_status(core_plugin, admin_context,
                                     subport.port_id, status)

    @log_helpers.log_method_call
    def trunk_subports_update_status(self, resource, event, trigger, payload):
        core_plugin = directory.get_plugin()
        admin_context = context.get_admin_context()
        port = payload.latest_state
        original_port = payload.states[0]
        if port['status'] == original_port['status']:
            return
        for subport_id in self._get_subports_ids(port['id']):
            self._set_subport_status(core_plugin, admin_context, subport_id,
                                     port['status'])

    def _set_subport_status(self, plugin, admin_context, port_id, status):
        plugin.update_port_status(admin_context, port_id, status)

    def _get_subports_ids(self, port_id):
        trunk_plugin = directory.get_plugin('trunk')
        filters = {'port_id': port_id}
        trunks = trunk_plugin.get_trunks(context.get_admin_context(),
                                         filters=filters)
        if not trunks:
            return ()
        trunk = trunks[0]
        return (subport['port_id'] for subport in trunk['sub_ports'])


@registry.has_registry_receivers
class OpenDaylightTrunkDriverV2(trunk_base.DriverBase):
    @property
    def is_loaded(self):
        try:
            return (odl_const.ODL_ML2_MECH_DRIVER_V2 in
                    cfg.CONF.ml2.mechanism_drivers)
        except cfg.NoSuchOptError:
            return False

    @registry.receives(resources.TRUNK_PLUGIN, [events.AFTER_INIT])
    def register(self, resource, event, trigger, payload=None):
        super(OpenDaylightTrunkDriverV2, self).register(
            resource, event, trigger, payload=payload)
        self._handler = OpenDaylightTrunkHandlerV2()
        registry.subscribe(self._handler.trunk_create_precommit,
                           resources.TRUNK, events.PRECOMMIT_CREATE)
        registry.subscribe(self._handler.trunk_create_postcommit,
                           resources.TRUNK, events.AFTER_CREATE)
        registry.subscribe(self._handler.trunk_update_precommit,
                           resources.TRUNK, events.PRECOMMIT_UPDATE)
        registry.subscribe(self._handler.trunk_update_postcommit,
                           resources.TRUNK, events.AFTER_UPDATE)
        registry.subscribe(self._handler.trunk_delete_precommit,
                           resources.TRUNK, events.PRECOMMIT_DELETE)
        registry.subscribe(self._handler.trunk_delete_postcommit,
                           resources.TRUNK, events.AFTER_DELETE)
        for event_ in (events.PRECOMMIT_CREATE, events.PRECOMMIT_DELETE):
            registry.subscribe(self._handler.trunk_update_precommit,
                               resources.SUBPORTS, event_)
        for event_ in (events.AFTER_CREATE, events.AFTER_DELETE):
            registry.subscribe(self._handler.trunk_update_postcommit,
                               resources.SUBPORTS, event_)
            # Upon subport creation/deletion we need to set the right port
            # status:
            # 1. Set it to parent status when it is attached to the trunk
            # 2. Set it to down when is removed from the trunk
            registry.subscribe(self._handler.trunk_subports_set_status,
                               resources.SUBPORTS, event_)
        # NOTE(ltomasbo): if the status of the parent port changes, the
        # subports need to update their status too
        registry.subscribe(self._handler.trunk_subports_update_status,
                           resources.PORT, events.AFTER_UPDATE)

    @classmethod
    def create(cls):
        return cls(odl_const.ODL_ML2_MECH_DRIVER_V2,
                   odltrunk_const.SUPPORTED_INTERFACES,
                   odltrunk_const.SUPPORTED_SEGMENTATION_TYPES,
                   None,
                   can_trunk_bound_port=True)



# Copyright (c) 2015 Intel Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
import io

import requests

from oslo_log import log as logging
from oslo_serialization import jsonutils

from networking_odl._i18n import _
from networking_odl.common import client
from networking_odl.common import constants as odl_const


LOG = logging.getLogger(__name__)

OK = requests.codes.ok
NO_CONTENT = requests.codes.no_content
NOT_ALLOWED = requests.codes.not_allowed
NOT_FOUND = requests.codes.not_found
BAD_REQUEST = requests.codes.bad_request


class OpenDaylightLwtClient(client.OpenDaylightRestClient):
    """Lightweight testing client"""

    lwt_dict = {odl_const.ODL_NETWORKS: {},
                odl_const.ODL_SUBNETS: {},
                odl_const.ODL_PORTS: {},
                odl_const.ODL_SGS: {},
                odl_const.ODL_SG_RULES: {}}

    @classmethod
    def _make_response(cls, status_code=OK, content=None):
        """Only supports 'content-type': 'application/json'"""
        response = requests.models.Response()
        response.status_code = status_code
        if content:
            response.raw = io.BytesIO(jsonutils.dumps(content).encode('utf-8'))

        return response

    @classmethod
    def _get_resource_id(cls, urlpath):
        # resource ID is the last element of urlpath
        return str(urlpath).rsplit('/', 1)[-1]

    @classmethod
    def post(cls, resource_type, resource_dict, urlpath, resource_list):
        """No ID in URL, elements in resource_list must have ID"""

        if resource_list is None:
            raise ValueError(_("resource_list can not be None"))

        for resource in resource_list:
            if resource['id'] in resource_dict:
                LOG.debug("%s %s already exists", resource_type,
                          resource['id'])
                response = cls._make_response(NOT_ALLOWED)
                raise requests.exceptions.HTTPError(response=response)

            resource_dict[resource['id']] = deepcopy(resource)

        return cls._make_response(NO_CONTENT)

    @classmethod
    def put(cls, resource_type, resource_dict, urlpath, resource_list):

        resource_id = cls._get_resource_id(urlpath)

        if resource_list is None:
            raise ValueError(_("resource_list can not be None"))

        if resource_id and len(resource_list) != 1:
            LOG.debug("Updating %s with multiple resources", urlpath)
            response = cls._make_response(BAD_REQUEST)
            raise requests.exceptions.HTTPError(response=response)

        for resource in resource_list:
            res_id = resource_id or resource['id']
            if res_id in resource_dict:
                resource_dict[res_id].update(deepcopy(resource))
            else:
                LOG.debug("%s %s does not exist", resource_type, res_id)
                response = cls._make_response(NOT_FOUND)
                raise requests.exceptions.HTTPError(response=response)

        return cls._make_response(NO_CONTENT)

    @classmethod
    def delete(cls, resource_type, resource_dict, urlpath, resource_list):

        if resource_list is None:
            resource_id = cls._get_resource_id(urlpath)
            id_list = [resource_id]
        else:
            id_list = [res['id'] for res in resource_list]

        for res_id in id_list:
            removed = resource_dict.pop(res_id, None)
            if removed is None:
                LOG.debug("%s %s does not exist", resource_type, res_id)
                response = cls._make_response(NOT_FOUND)
                raise requests.exceptions.HTTPError(response=response)

        return cls._make_response(NO_CONTENT)

    @classmethod
    def get(cls, resource_type, resource_dict, urlpath, resource_list=None):

        resource_id = cls._get_resource_id(urlpath)

        if resource_id:
            resource = resource_dict.get(resource_id)
            if resource is None:
                LOG.debug("%s %s does not exist", resource_type, resource_id)
                response = cls._make_response(NOT_FOUND)
                raise requests.exceptions.HTTPError(response=response)

            # When getting single resource, return value is a dict
            r_list = {resource_type[:-1]: deepcopy(resource)}
            return cls._make_response(OK, r_list)

        r_list = [{resource_type[:-1]: deepcopy(res)}
                  for res in resource_dict.values()]

        return cls._make_response(OK, r_list)

    def sendjson(self, method, urlpath, obj=None):
        """Lightweight testing without ODL"""

        if '/' not in urlpath:
            urlpath += '/'

        resource_type = str(urlpath).split('/', 1)[0]
        resource_type = resource_type.replace('-', '_')

        resource_dict = self.lwt_dict.get(resource_type)

        if resource_dict is None:
            LOG.debug("Resource type %s is not supported", resource_type)
            response = self._make_response(NOT_FOUND)
            raise requests.exceptions.HTTPError(response=response)

        func = getattr(self, str(method).lower())

        resource_list = None
        if obj:
            # If obj is not None, it can only have one entry
            assert len(obj) == 1, "Obj can only have one entry"

            key, resource_list = list(obj.items())[0]

            if not isinstance(resource_list, list):
                # Need to transform resource_list to a real list, i.e. [res]
                resource_list = [resource_list]

        return func(resource_type, resource_dict, urlpath, resource_list)



# Copyright (c) 2014 Red Hat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import urllib

from oslo_config import cfg

from networking_odl.common import constants as odl_const

cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')


def try_del(d, keys):
    """Ignore key errors when deleting from a dictionary."""
    for key in keys:
        try:
            del d[key]
        except KeyError:
            pass


def make_url_object(object_type):
    obj_pl = odl_const.RESOURCE_URL_MAPPINGS.get(object_type, None)
    if obj_pl is None:
        obj_pl = neutronify(object_type + 's')
    return obj_pl


# TODO(manjeets) consolidate this method with make_url_object
def neutronify(name):
    """Adjust the resource name for use with Neutron's API"""
    return name.replace('_', '-')


def get_odl_url(path=''):
    '''Make a URL for some ODL resource (path)'''
    purl = urllib.parse.urlsplit(cfg.CONF.ml2_odl.url)
    features_url = urllib.parse.urlunparse((
        purl.scheme, purl.netloc, path, '', '', ''))
    return features_url



# Copyright 2018 Intel Corporation.
# Copyright 2018 Isaku Yamahata <isaku.yamahata at intel com>
#                               <isaku.yamahata at gmail com>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from neutron.objects import router as l3_obj
from neutron.services.l3_router.service_providers import base
from neutron_lib.callbacks import events
from neutron_lib.callbacks import priority_group
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as q_const
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_odl.common import constants as odl_const
from networking_odl.journal import full_sync
from networking_odl.journal import journal


LOG = logging.getLogger(__name__)

L3_RESOURCES = {
    odl_const.ODL_ROUTER: odl_const.ODL_ROUTERS,
    odl_const.ODL_FLOATINGIP: odl_const.ODL_FLOATINGIPS
}


@registry.has_registry_receivers
class ODLL3ServiceProvider(base.L3ServiceProvider):
    @log_helpers.log_method_call
    def __init__(self, l3_plugin):
        super(ODLL3ServiceProvider, self).__init__(l3_plugin)
        self.journal = journal.OpenDaylightJournalThread()
        # TODO(yamahata): add method for fullsync to retrieve
        # all the router with odl service provider.
        # other router with other service provider should be filtered.
        full_sync.register(plugin_constants.L3, L3_RESOURCES)
        self.odl_provider = __name__ + "." + self.__class__.__name__

    @property
    def _flavor_plugin(self):
        try:
            return self._flavor_plugin_ref
        except AttributeError:
            self._flavor_plugin_ref = directory.get_plugin(
                plugin_constants.FLAVORS)
            return self._flavor_plugin_ref

    def _validate_l3_flavor(self, context, router_id):
        if router_id is None:
            return False
        router = l3_obj.Router.get_object(context, id=router_id)
        flavor = self._flavor_plugin.get_flavor(context, router.flavor_id)
        provider = self._flavor_plugin.get_flavor_next_provider(
            context, flavor['id'])[0]
        return str(provider['driver']) == self.odl_provider

    def _update_floatingip_status(self, context, fip_dict):
        port_id = fip_dict.get('port_id')
        status = q_const.ACTIVE if port_id else q_const.DOWN
        l3_obj.FloatingIP.update_object(context, {'status': status},
                                        id=fip_dict['id'])

    @registry.receives(resources.ROUTER_CONTROLLER,
                       [events.PRECOMMIT_ADD_ASSOCIATION])
    @log_helpers.log_method_call
    def _router_add_association(self, resource, event, trigger, payload=None):
        context = payload.context
        router_dict = payload.request_body
        router_dict['gw_port_id'] = payload.latest_state.gw_port_id
        router_id = payload.resource_id
        if not self._validate_l3_flavor(context, router_id):
            return
        journal.record(context, odl_const.ODL_ROUTER, router_dict['id'],
                       odl_const.ODL_CREATE, router_dict)

    @registry.receives(resources.ROUTER, [events.PRECOMMIT_UPDATE],
                       priority_group.PRIORITY_ROUTER_DRIVER)
    @log_helpers.log_method_call
    def _router_update_precommit(self, resource, event, trigger, **kwargs):
        # NOTE(manjeets) router update bypasses the driver controller
        # and argument type is different.
        payload = kwargs.get('payload', None)
        if payload:
            context = payload.context
            router_id = payload.states[0]['id']
            router_dict = payload.request_body
            gw_port_id = payload.states[0]['gw_port_id']
        else:
            # TODO(manjeets) Remove this shim once payload is fully adapted
            # https://bugs.launchpad.net/neutron/+bug/1747747
            context = kwargs['context']
            router_id = kwargs['router_db'].id
            router_dict = kwargs['router']
            gw_port_id = kwargs['router_db'].gw_port_id
        if not self._validate_l3_flavor(context, router_id):
            return
        if 'gw_port_id' not in router_dict:
            router_dict['gw_port_id'] = gw_port_id
        journal.record(context, odl_const.ODL_ROUTER,
                       router_id, odl_const.ODL_UPDATE, router_dict)

    @registry.receives(resources.ROUTER_CONTROLLER,
                       [events.PRECOMMIT_DELETE_ASSOCIATIONS])
    @log_helpers.log_method_call
    def _router_del_association(self, resource, event, trigger, payload=None):
        router_id = payload.latest_state.id
        context = payload.context
        if not self._validate_l3_flavor(context, router_id):
            return
        # TODO(yamahata): process floating ip etc. or just raise error?
        dependency_list = [payload.latest_state.gw_port_id]
        journal.record(context, odl_const.ODL_ROUTER, router_id,
                       odl_const.ODL_DELETE, dependency_list)

    @registry.receives(resources.FLOATING_IP, [events.PRECOMMIT_CREATE])
    @log_helpers.log_method_call
    def _floatingip_create_precommit(self, resource, event, trigger, payload):
        context = payload.context
        fip_dict = copy.deepcopy(payload.states[0])
        router_id = payload.desired_state.router_id
        if not self._validate_l3_flavor(context, router_id):
            return
        fip_dict['id'] = payload.resource_id
        self._update_floatingip_status(context, fip_dict)
        if fip_dict['floating_ip_address'] is None:
            fip_dict['floating_ip_address'] = \
                payload.desired_state.floating_ip_address
        journal.record(context, odl_const.ODL_FLOATINGIP, fip_dict['id'],
                       odl_const.ODL_CREATE, fip_dict)

    @registry.receives(resources.FLOATING_IP, [events.PRECOMMIT_UPDATE])
    @log_helpers.log_method_call
    def _floatingip_update_precommit(self, resource, event, trigger, payload):
        context = payload.context
        fip_dict = payload.states[0]
        router_id = payload.desired_state.router_id
        fip_dict['id'] = payload.desired_state.id
        if not self._validate_l3_flavor(context, router_id):
            return
        self._update_floatingip_status(context, fip_dict)
        journal.record(context, odl_const.ODL_FLOATINGIP, fip_dict['id'],
                       odl_const.ODL_UPDATE, fip_dict)

    @registry.receives(resources.FLOATING_IP, [events.PRECOMMIT_DELETE])
    @log_helpers.log_method_call
    def _floatingip_delete_precommit(self, resource, event, trigger, payload):
        context = payload.context
        floating_port_id = payload.resource_id
        fip_data = l3_obj.FloatingIP.get_objects(
            context,
            floating_port_id=floating_port_id)[0]
        if not self._validate_l3_flavor(context, fip_data.router_id):
            return
        dependency_list = [fip_data.router_id, fip_data.floating_network_id]
        journal.record(context, odl_const.ODL_FLOATINGIP, fip_data.id,
                       odl_const.ODL_DELETE, dependency_list)

    @registry.receives(resources.FLOATING_IP, [events.AFTER_CREATE,
                                               events.AFTER_UPDATE,
                                               events.AFTER_DELETE])
    @registry.receives(resources.ROUTER, [events.AFTER_CREATE,
                                          events.AFTER_UPDATE,
                                          events.AFTER_DELETE])
    @log_helpers.log_method_call
    def _l3_postcommit(self, resource, event, trigger, payload):
        self.journal.set_sync_event()



#!/usr/bin/env python

# Copyright (c) 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""
Command line script to analyze journal entry processing time based on logs.
By default the input is read through pipe, unless a log file is specified.

Examples:
    Analyzing devstack's Neutron log:
        journalctl -u devstack@neutron-api | python analyze_journal.py

    Analyzing an arbitrary log file:
        python analyze_journal.py --file /path/to/file.log
"""

import collections
import re
import sys

from oslo_config import cfg

from networking_odl._i18n import _
from networking_odl.journal import journal

COMMAND_LINE_OPTIONS = [
    cfg.StrOpt('file', default=None,
               help=_("Log file to analyze.")),
    cfg.IntOpt('slowest', min=1, default=10,
               help=_("Prints the N slowest entries (10 by default).")),
]

# This regex will match any replacement key in the log message and extract
# the key name.
KEY_MATCHER = re.compile(r'\%\((\S+)\)s')
LOG_KEYS = KEY_MATCHER.findall(journal.LOG_ENTRY_TEMPLATE)
KEY_TEMP_PATTERN = 'KEYPATTERN'
LOG_MATCHER = re.compile(
    re.sub(KEY_TEMP_PATTERN, r'(\\S+)', re.escape(
        KEY_MATCHER.sub(KEY_TEMP_PATTERN, journal.LOG_ENTRY_TEMPLATE))))
ENTRY_LOG_TEMPLATE = ' * Entry id: %s, processing time: %.3fs; %s %s %s'

EntryStats = collections.namedtuple(
    'EntryStats', 'entry_id time op obj_type obj_id')


def setup_conf(output, args):
    """setup cmdline options."""

    if any(flag in args for flag in ('-h', '--help')):
        print(__doc__, file=output)

    conf = cfg.ConfigOpts()
    conf.register_cli_opts(COMMAND_LINE_OPTIONS)
    conf(args=args)
    return conf


def parse_log(content):
    entries = {}
    for line in content:
        matched = LOG_MATCHER.search(line)
        if matched is None:
            continue

        entry_log = dict(zip(LOG_KEYS, matched.groups()))
        entry_id = entry_log['entry_id']
        entry = entries.get(entry_id, entry_log)

        log_type = entry_log['log_type']
        entry[log_type] = float(entry_log['timestamp'])
        entries[entry_id] = entry

    return entries


def analyze_entries(entries):
    entries_stats = []
    for entry_id, entry in entries.items():
        recorded_time = entry.get(journal.LOG_RECORDED, None)
        completed_time = entry.get(journal.LOG_COMPLETED, None)
        if recorded_time is None or completed_time is None:
            continue

        delta = completed_time - recorded_time
        entries_stats.append(EntryStats(
            entry_id=entry_id, time=delta, op=entry['op'],
            obj_type=entry['obj_type'], obj_id=entry['obj_id']))

    return entries_stats


def _percentile(timings, percent):
    location = int(len(timings) * (percent / 100.0))
    return int(timings[location])


def print_stats(output, slowest, entries_stats):
    entries_stats = sorted(
        entries_stats, key=lambda entry_stats: entry_stats.time)
    timings = [entry_stats.time for entry_stats in entries_stats]
    avg = sum(timings) / len(timings)

    print('Average processing time: %ss' % avg, file=output)
    print('90th percentile: %ss' % _percentile(timings, 90), file=output)
    print('99th percentile: %ss' % _percentile(timings, 99), file=output)
    print('99.9th percentile: %ss' % _percentile(timings, 99.9), file=output)
    print('%s slowest entries:' % slowest, file=output)
    slowest = entries_stats[:-(slowest + 1):-1]
    for entry_stats in slowest:
        print(ENTRY_LOG_TEMPLATE % entry_stats, file=output)


def get_content(file_name):
    return open(file_name, encoding="utf8") if file_name else sys.stdin


def main(output=sys.stdout):
    conf = setup_conf(output, sys.argv[1:])
    with get_content(conf.file) as content:
        entries = parse_log(content)

    entries_stats = analyze_entries(entries)

    if not entries_stats:
        print('No entry statistics found.', file=output)
        return 1

    print_stats(output, conf.slowest, entries_stats)

    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2016 Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy


def enforce_policy_format(policy):
    if 'bandwidth_limit_rules' not in policy.keys():
        policy['bandwidth_limit_rules'] = []
    if 'dscp_marking_rules' not in policy.keys():
        policy['dscp_marking_rules'] = []
    return policy


# NOTE(manjeets) keeping common methods for formatting
# qos data in qos_utils for code reuse.
def convert_rules_format(data):
    policy = copy.deepcopy(data)
    policy.pop('tenant_id', None)
    policy.pop('rules', None)
    for rule in data.get('rules', []):
        rule_type = rule['type'] + '_rules'
        rule.pop('type', None)
        rule.pop('qos_policy_id', None)
        rule['tenant_id'] = data['tenant_id']
        policy[rule_type] = [rule]
    return enforce_policy_format(policy)



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from datetime import datetime
import threading
import time

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib import context as nl_context
from neutron_lib.db import api as db_api
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_db import exception
from oslo_log import log as logging
from requests import exceptions

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.common import filters
from networking_odl.common import utils
from networking_odl.db import db
from networking_odl.journal import dependency_validations


LOG = logging.getLogger(__name__)

MAKE_URL = {}
LOG_ENTRY_TEMPLATE = ("%(log_type)s (Entry ID: %(entry_id)s) - %(op)s "
                      "%(obj_type)s %(obj_id)s (Time stamp: %(timestamp)s)")
LOG_RECORDED = 'Recorded'
LOG_PROCESSING = 'Processing'
LOG_COMPLETED = 'Completed'
LOG_ERROR_PROCESSING = 'Error while processing'


def call_thread_on_end(func):
    def new_func(obj, *args, **kwargs):
        return_value = func(obj, *args, **kwargs)
        obj.journal.set_sync_event()
        return return_value
    return new_func


def _enrich_port(plugin_context, ml2_context, object_type, operation, data):
    """Enrich the port with additional information needed by ODL"""

    # NOTE(yamahata): work around of ODL neutron northbound
    # It passes security groups in port as list of dict for historical reasons.
    # keep its format for compatibility.
    # TODO(yamahata): drop this format conversion.
    if data[odl_const.ODL_SGS]:
        groups = [{'id': id_} for id_ in data['security_groups']]
    else:
        groups = []
    new_data = copy.deepcopy(data)
    new_data[odl_const.ODL_SGS] = groups

    # NOTE(yamahata): work around for port creation for router
    # tenant_id=''(empty string) is passed when port is created
    # by l3 plugin internally for router.
    # On the other hand, ODL doesn't accept empty string for tenant_id.
    # In that case, deduce tenant_id from network_id for now.
    # Right fix: modify Neutron so that don't allow empty string
    # for tenant_id even for port for internal use.
    # TODO(yamahata): eliminate this work around when neutron side
    # is fixed
    # assert port['tenant_id'] != ''
    if ('tenant_id' not in new_data or new_data['tenant_id'] == ''):
        if ml2_context:
            network = ml2_context._network_context._network
        else:
            plugin = directory.get_plugin()
            network = plugin.get_network(plugin_context,
                                         new_data['network_id'])
        new_data['tenant_id'] = network['tenant_id']

    return new_data


def _log_entry(log_type, entry, log_level=logging.INFO, **kwargs):
    delta = datetime.now() - datetime.min
    timestamp = delta.total_seconds()
    log_dict = {'log_type': log_type, 'op': entry.operation,
                'obj_type': entry.object_type, 'obj_id': entry.object_uuid,
                'entry_id': entry.seqnum, 'timestamp': timestamp}
    LOG.log(log_level, LOG_ENTRY_TEMPLATE, log_dict, **kwargs)


def record(plugin_context, object_type, object_uuid, operation, data,
           ml2_context=None):
    if (object_type == odl_const.ODL_PORT and
            operation in (odl_const.ODL_CREATE, odl_const.ODL_UPDATE)):
        data = _enrich_port(
            plugin_context, ml2_context, object_type, operation, data)

    # Calculate depending_on on other journal entries
    depending_on = dependency_validations.calculate(
        plugin_context, operation, object_type, object_uuid, data)

    # NOTE(mpeterson): Between the moment that a dependency is calculated and
    # the new entry is recorded in the journal, an operation can ocurr that
    # would make the dependency irrelevant. In that case we request a retry.
    # For more details, read the commit message that introduced this comment.
    try:
        entry = db.create_pending_row(
            plugin_context, object_type, object_uuid, operation, data,
            depending_on=depending_on)
    except exception.DBReferenceError as e:
        raise exception.RetryRequest(e)

    _log_entry(LOG_RECORDED, entry)
    LOG.debug('Entry with ID %(entry_id)s depends on these entries: '
              '%(depending_on)s',
              {'entry_id': entry.seqnum,
               'depending_on': [d.seqnum for d in depending_on]})


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_WRITER.savepoint
def entry_complete(context, entry):
    if cfg.CONF.ml2_odl.completed_rows_retention == 0:
        db.delete_row(context, entry)
    else:
        db.update_db_row_state(context, entry, odl_const.COMPLETED)
        db.delete_dependency(context, entry)


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_WRITER.savepoint
def entry_reset(context, entry):
    db.update_db_row_state(context, entry, odl_const.PENDING)


@db_api.retry_if_session_inactive()
@db_api.CONTEXT_WRITER.savepoint
def entry_update_state_by_retry_count(context, entry, retry_count):
    db.update_pending_db_row_retry(context, entry, retry_count)


def _make_url(row):
    url_object = utils.make_url_object(row.object_type)
    urlpath = ''
    if row.operation == odl_const.ODL_CREATE:
        urlpath = url_object
    else:
        urlpath = url_object + '/' + row.object_uuid

    return urlpath


def register_url_builder(object_type, method):
    MAKE_URL[object_type] = method


def _build_url(row):
    return MAKE_URL.get(row.object_type, _make_url)(row)


class OpenDaylightJournalThread(object):
    """Thread worker for the OpenDaylight Journal Database."""

    # make those parameter configurable?
    _RETRY_SLEEP_MIN = 0.1
    _RETRY_SLEEP_MAX = 60

    def __init__(self, start_thread=True):
        self.client = client.OpenDaylightRestClient.create_client()
        self._max_retry_count = cfg.CONF.ml2_odl.retry_count
        self._sleep_time = self._RETRY_SLEEP_MIN
        self.event = threading.Event()
        self._odl_sync_thread = self._create_odl_sync_thread()
        self._odl_sync_thread_stop = threading.Event()
        if start_thread:
            self.start()

    def _create_odl_sync_thread(self):
        return threading.Thread(name='sync', target=self.run_sync_thread)

    def start(self):
        # Start the sync thread
        LOG.debug("Starting a new sync thread")
        if self._odl_sync_thread_stop.is_set():
            self._odl_sync_thread_stop.clear()
            self._odl_sync_thread = self._create_odl_sync_thread()

        if not self._odl_sync_thread.is_alive():
            self._odl_sync_thread.start()

    def stop(self, timeout=None):
        """Allows to stop the sync thread.

        Args:
            timeout (float): Time in seconds to wait for joining or None for
                             no timeout.
        """
        # Stop the sync thread
        LOG.debug("Stopping the sync thread")
        if self._odl_sync_thread.is_alive():
            self._odl_sync_thread_stop.set()
            # Process the journal one last time before stopping.
            self.set_sync_event()
            self._odl_sync_thread.join(timeout)

    def set_sync_event(self):
        self.event.set()

    @staticmethod
    def _json_data(row):
        data = copy.deepcopy(row.data)
        filters.filter_for_odl(row.object_type, row.operation, data)

        if row.operation == odl_const.ODL_CREATE:
            method = 'post'
            to_send = {row.object_type: data}
        elif row.operation == odl_const.ODL_UPDATE:
            method = 'put'
            to_send = {row.object_type: data}
        elif row.operation == odl_const.ODL_DELETE:
            method = 'delete'
            to_send = None

        return method, _build_url(row), to_send

    def run_sync_thread(self):
        while not self._odl_sync_thread_stop.is_set():
            try:
                self.event.wait()
                self.event.clear()

                self.sync_pending_entries()
            except Exception:
                # Catch exceptions to protect the thread while running
                LOG.exception("Error on run_sync_thread")

    def sync_pending_entries(self):
        LOG.debug("Start processing journal entries")
        context = nl_context.get_admin_context()
        entry = db.get_oldest_pending_db_row_with_lock(context)
        if entry is None:
            LOG.debug("No journal entries to process")
            return

        while entry is not None:
            stop_processing = self._sync_entry(context, entry)
            if stop_processing:
                break

            entry = db.get_oldest_pending_db_row_with_lock(context)
        LOG.debug("Finished processing journal entries")

    def _retry_sleep(self):
        # When something happened in the connection to ODL, don't busy loop
        # because it's likely to hit same issue.
        # Wait for a while for recovery
        time.sleep(self._sleep_time)
        self._sleep_time = min(self._sleep_time * 2, self._RETRY_SLEEP_MAX)

    def _retry_reset(self):
        self._sleep_time = self._RETRY_SLEEP_MIN

    def _sync_entry(self, context, entry):
        _log_entry(LOG_PROCESSING, entry)
        method, urlpath, to_send = self._json_data(entry)

        # TODO(mkolesni): This logic is weirdly written, need to refactor it.
        try:
            self.client.sendjson(method, urlpath, to_send)
            registry.publish(entry.object_type, odl_const.BEFORE_COMPLETE,
                             self,
                             payload=events.DBEventPayload(
                                 context,
                                 metadata={'operation': entry.operation,
                                           'row': entry}))
            entry_complete(context, entry)
            self._retry_reset()
            _log_entry(LOG_COMPLETED, entry)
        except exceptions.ConnectionError:
            # Don't raise the retry count, just log an error & break
            entry_reset(context, entry)
            LOG.error("Cannot connect to the OpenDaylight Controller,"
                      " will not process additional entries")
            self._retry_sleep()
            return True
        except Exception:
            _log_entry(LOG_ERROR_PROCESSING, entry,
                       log_level=logging.ERROR, exc_info=True)
            entry_update_state_by_retry_count(
                context, entry, self._max_retry_count)

        return False



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests import base
from oslo_config import fixture as config_fixture

from networking_odl.common import constants as odl_const
from networking_odl.common import utils


class TestUtils(base.DietTestCase):

    def setUp(self):
        self.cfg = self.useFixture(config_fixture.Config())
        super(TestUtils, self).setUp()

    # TODO(manjeets) remove this test once neutronify is
    # consolidated with make_plural
    def test_neutronify(self):
        self.assertEqual('a-b-c', utils.neutronify('a_b_c'))

    def test_neutronify_empty(self):
        self.assertEqual('', utils.neutronify(''))

    @staticmethod
    def _get_resources():
        return {odl_const.ODL_SG: 'security-groups',
                odl_const.ODL_SG_RULE: 'security-group-rules',
                odl_const.ODL_NETWORK: 'networks',
                odl_const.ODL_SUBNET: 'subnets',
                odl_const.ODL_ROUTER: 'routers',
                odl_const.ODL_PORT: 'ports',
                odl_const.ODL_FLOATINGIP: 'floatingips',
                odl_const.ODL_QOS_POLICY: 'qos/policies',
                odl_const.ODL_TRUNK: 'trunks',
                odl_const.ODL_BGPVPN: 'bgpvpns',
                odl_const.ODL_SFC_FLOW_CLASSIFIER: 'sfc/flowclassifiers',
                odl_const.ODL_SFC_PORT_PAIR: 'sfc/portpairs',
                odl_const.ODL_SFC_PORT_PAIR_GROUP: 'sfc/portpairgroups',
                odl_const.ODL_SFC_PORT_CHAIN: 'sfc/portchains',
                odl_const.ODL_L2GATEWAY: 'l2-gateways',
                odl_const.ODL_L2GATEWAY_CONNECTION: 'l2gateway-connections'}

    def test_all_resources_url(self):
        for obj, url in self._get_resources().items():
            self.assertEqual(utils.make_url_object(obj), url)

    def test_get_odl_url(self):
        """test make uri."""
        self.cfg.config(url='http://localhost:8080/controller/nb/v2/neutron',
                        group='ml2_odl')
        test_path = '/restconf/neutron:neutron/hostconfigs'
        expected = "http://localhost:8080/restconf/neutron:neutron/hostconfigs"
        test_uri = utils.get_odl_url(path=test_path)

        self.assertEqual(expected, test_uri)



# Copyright (c) 2015 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy
import datetime
import operator
from unittest import mock

import requests
import testscenarios

from neutron.db.models import securitygroup
from neutron.db import segments_db
from neutron.plugins.ml2 import plugin
from neutron.tests.unit import testlib_api
from neutron_lib.api.definitions import multiprovidernet as mpnet_apidef
from neutron_lib.api.definitions import provider_net as providernet
from neutron_lib.callbacks import events
from neutron_lib import constants as n_constants
from neutron_lib.db import api as db_api
from neutron_lib.plugins import constants as plugin_const
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from networking_odl.common import callback
from networking_odl.common import constants as odl_const
from networking_odl.common import filters
from networking_odl.common import utils
from networking_odl.db import db
from networking_odl.journal import base_driver
from networking_odl.journal import cleanup
from networking_odl.journal import journal
from networking_odl.ml2 import mech_driver_v2
from networking_odl.tests import base
from networking_odl.tests.unit import base_v2


# Required to generate tests from scenarios. Not compatible with nose.
load_tests = testscenarios.load_tests_apply_scenarios

cfg.CONF.import_group('ml2_odl', 'networking_odl.common.config')

SECURITY_GROUP = '2f9244b4-9bee-4e81-bc4a-3f3c2045b3d7'
SG_FAKE_ID = uuidutils.generate_uuid()
SG_RULE_FAKE_ID = uuidutils.generate_uuid()


class OpenDayLightMechanismConfigTests(testlib_api.SqlTestCase):
    def setUp(self):
        self.useFixture(base.OpenDaylightFeaturesFixture())
        self.useFixture(base.OpenDaylightJournalThreadFixture())
        self.useFixture(base.OpenDaylightPseudoAgentPrePopulateFixture())
        self.cfg = self.useFixture(config_fixture.Config())
        super(OpenDayLightMechanismConfigTests, self).setUp()
        self.cfg.config(mechanism_drivers=[
                        'logger', 'opendaylight_v2'], group='ml2')
        self.cfg.config(
            port_binding_controller='legacy-port-binding', group='ml2_odl')

    def _set_config(self, url='http://127.0.0.1:9999', username='someuser',
                    password='somepass'):
        self.cfg.config(url=url, group='ml2_odl')
        self.cfg.config(username=username, group='ml2_odl')
        self.cfg.config(password=password, group='ml2_odl')

    def _test_missing_config(self, **kwargs):
        self._set_config(**kwargs)
        self.assertRaisesRegex(cfg.RequiredOptError,
                               r'value required for option \w+ in group '
                               r'\[ml2_odl\]',
                               plugin.Ml2Plugin)

    def test_valid_config(self):
        self._set_config()
        plugin.Ml2Plugin()

    def test_missing_url_raises_exception(self):
        self._test_missing_config(url=None)

    def test_missing_username_raises_exception(self):
        self._test_missing_config(username=None)

    def test_missing_password_raises_exception(self):
        self._test_missing_config(password=None)


class _OpenDaylightMechanismBase(base_v2.OpenDaylightTestCase):
    _mechanism_drivers = ['logger', 'opendaylight_v2']
    # TODO(mpeterson): Add a test to make sure extension_drivers are honored.
    _extension_drivers = ['port_security', 'qos']

    def setUp(self):
        mech_initialize_patcher = mock.patch.object(
            mech_driver_v2.OpenDaylightMechanismDriver,
            'initialize',
            autospec=True,
            side_effect=mech_driver_v2.OpenDaylightMechanismDriver.initialize
        )
        self.mech_initialize_mock = mech_initialize_patcher.start()
        mock.patch('networking_odl.common.odl_features.init').start()
        # NOTE(mpeterson): We cannot use stop in the following cleanup because
        # several of the following fixtures and setUp() add a cleanup for
        # stopall. The reason to add the stopall ourselves is to make sure
        # that it will be stopped if anything were to change in the future.
        self.addCleanup(mock.patch.stopall)
        self.useFixture(base.OpenDaylightPseudoAgentPrePopulateFixture())
        self.cfg = self.useFixture(config_fixture.Config())
        self.cfg.config(extension_drivers=self._extension_drivers, group='ml2')
        super(_OpenDaylightMechanismBase, self).setUp()

    def test_mechanism_driver_is_initialized(self):
        """Test that the mech driver is initialized.

        This test will allow us know if the mech driver is not initialized
        in case there is a change in the way Ml2PluginV2TestCase instantiate
        them
        """
        # NOTE(mpeterson): Because of the autospec the mock lacks
        # the helper assert_called_once
        msg = "The opendaylight_v2 ML2 Mechanism Driver was not initialized"
        self.assertTrue(self.mech_initialize_mock.called, msg)


class DataMatcher(object):

    def __init__(self, operation, object_type, context):
        if object_type in [odl_const.ODL_SG, odl_const.ODL_SG_RULE]:
            self._data = copy.deepcopy(context[object_type])
        elif object_type == odl_const.ODL_PORT:
            # NOTE(yamahata): work around for journal._enrich_port()
            self._data = copy.deepcopy(context.current)
            if self._data.get(odl_const.ODL_SGS):
                self._data[odl_const.ODL_SGS] = [
                    {'id': id_} for id_ in self._data[odl_const.ODL_SGS]]
        else:
            self._data = copy.deepcopy(context.current)
        self._object_type = object_type
        filters.filter_for_odl(object_type, operation, self._data)

    def __eq__(self, s):
        data = jsonutils.loads(s)
        return self._data == data[self._object_type]

    def __ne__(self, s):
        return not self.__eq__(s)

    def __repr__(self):
        # for debugging
        return 'DataMatcher(%(object_type)s, %(data)s)' % {
            'object_type': self._object_type,
            'data': self._data}


class AttributeDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttributeDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


class OpenDaylightMechanismDriverTestCase(base_v2.OpenDaylightConfigBase):
    def setUp(self):
        self.useFixture(base.OpenDaylightFeaturesFixture())
        self.useFixture(base.OpenDaylightJournalThreadFixture())
        self.useFixture(base.OpenDaylightPseudoAgentPrePopulateFixture())
        super(OpenDaylightMechanismDriverTestCase, self).setUp()
        self.mech = mech_driver_v2.OpenDaylightMechanismDriver()
        self.mech.initialize()

    def test_registered_plugin_type(self):
        self.assertEqual(self.mech.plugin_type, plugin_const.CORE)

    def test_registered_resources(self):
        for resource in self.mech.RESOURCES:
            self.assertIn(resource, base_driver.ALL_RESOURCES)
            self.assertEqual(base_driver.ALL_RESOURCES[resource], self.mech)

    def _get_mock_network_operation_context(self):
        current = {'status': 'ACTIVE',
                   'subnets': [],
                   'name': 'net1',
                   'provider:physical_network': None,
                   'admin_state_up': True,
                   'tenant_id': 'test-tenant',
                   'provider:network_type': 'local',
                   'router:external': False,
                   'shared': False,
                   'id': 'd897e21a-dfd6-4331-a5dd-7524fa421c3e',
                   'provider:segmentation_id': None}
        context = mock.Mock(current=current)
        context._plugin_context = self.db_context
        return context

    def _get_mock_subnet_operation_context(self):
        current = {'ipv6_ra_mode': None,
                   'allocation_pools': [{'start': '10.0.0.2',
                                         'end': '10.0.1.254'}],
                   'host_routes': [],
                   'ipv6_address_mode': None,
                   'cidr': '10.0.0.0/23',
                   'id': '72c56c48-e9b8-4dcf-b3a7-0813bb3bd839',
                   'name': '',
                   'enable_dhcp': True,
                   'network_id': 'd897e21a-dfd6-4331-a5dd-7524fa421c3e',
                   'tenant_id': 'test-tenant',
                   'dns_nameservers': [],
                   'gateway_ip': '10.0.0.1',
                   'ip_version': 4,
                   'shared': False}
        context = mock.Mock(current=current)
        context._plugin_context = self.db_context
        return context

    def _get_mock_port_operation_context(self):
        current = {'status': 'DOWN',
                   'binding:host_id': '',
                   'allowed_address_pairs': [],
                   'device_owner': 'fake_owner',
                   'binding:profile': {},
                   'fixed_ips': [{
                       'subnet_id': '72c56c48-e9b8-4dcf-b3a7-0813bb3bd839'}],
                   'id': '83d56c48-e9b8-4dcf-b3a7-0813bb3bd940',
                   'security_groups': [SECURITY_GROUP],
                   'device_id': 'fake_device',
                   'name': '',
                   'admin_state_up': True,
                   'network_id': 'd897e21a-dfd6-4331-a5dd-7524fa421c3e',
                   'tenant_id': 'test-tenant',
                   'binding:vif_details': {},
                   'binding:vnic_type': 'normal',
                   'binding:vif_type': 'unbound',
                   'mac_address': '12:34:56:78:21:b6'}
        _network = self._get_mock_network_operation_context().current
        _plugin = directory.get_plugin()
        _plugin.writer_get_security_group = mock.Mock(
            return_value=SECURITY_GROUP)
        _plugin.get_port = mock.Mock(return_value=current)
        _plugin.get_network = mock.Mock(return_value=_network)
        _plugin_context_mock = {'session': self.db_context.session}
        _network_context_mock = {'_network': _network}
        context = {'current': AttributeDict(current),
                   '_plugin': _plugin,
                   '_plugin_context': AttributeDict(_plugin_context_mock),
                   '_network_context': AttributeDict(_network_context_mock)}
        return AttributeDict(context)

    def _get_mock_security_group_operation_context(self):
        context = {odl_const.ODL_SG: {'name': 'test_sg',
                                      'project_id': 'test-tenant',
                                      'tenant_id': 'test-tenant',
                                      'description': 'test-description',
                                      'security_group_rules': [],
                                      'id': SG_FAKE_ID}}
        return context

    def _get_mock_security_group_rule_operation_context(self):
        context = {odl_const.ODL_SG_RULE: {'security_group_id': SG_FAKE_ID,
                                           'id': SG_RULE_FAKE_ID}}
        _plugin = directory.get_plugin()
        _plugin._get_security_group_rule = mock.Mock(
            return_value=AttributeDict(context[odl_const.ODL_SG_RULE]))
        return context

    def _get_mock_operation_context(self, object_type):
        getter = getattr(self, '_get_mock_%s_operation_context' % object_type)
        return getter()

    _status_code_msgs = {
        200: '',
        201: '',
        204: '',
        400: '400 Client Error: Bad Request',
        401: '401 Client Error: Unauthorized',
        403: '403 Client Error: Forbidden',
        404: '404 Client Error: Not Found',
        409: '409 Client Error: Conflict',
        501: '501 Server Error: Not Implemented',
        503: '503 Server Error: Service Unavailable',
    }

    @classmethod
    def _get_mock_request_response(cls, status_code):
        response = mock.Mock(status_code=status_code)
        response.raise_for_status = mock.Mock() if status_code < 400 else (
            mock.Mock(side_effect=requests.exceptions.HTTPError(
                cls._status_code_msgs[status_code])))
        return response

    def _test_operation(self, status_code, expected_calls,
                        *args, **kwargs):
        request_response = self._get_mock_request_response(status_code)
        with mock.patch('requests.sessions.Session.request',
                        return_value=request_response) as mock_method:
            self.run_journal_processing()

        if expected_calls:
            mock_method.assert_called_with(
                headers={'Content-Type': 'application/json'},
                timeout=cfg.CONF.ml2_odl.timeout, *args, **kwargs)
        self.assertEqual(expected_calls, mock_method.call_count)

    def _call_operation_object(self, operation, object_type):
        context = self._get_mock_operation_context(object_type)
        if object_type in [odl_const.ODL_SG, odl_const.ODL_SG_RULE]:
            res_type = [rt for rt in callback._RESOURCE_MAPPING.values()
                        if rt.singular == object_type][0]
            res_id = context[object_type]['id']
            context_ = (copy.deepcopy(context)
                        if operation != odl_const.ODL_DELETE else None)
            plugin_context = self.db_context
            if (object_type == odl_const.ODL_SG and
                    operation in [
                        odl_const.ODL_CREATE,
                        odl_const.ODL_DELETE]):
                # TODO(yamahata): remove this work around once
                # https://review.opendev.org/#/c/281693/
                # is merged.
                if operation == odl_const.ODL_CREATE:
                    sg = securitygroup.SecurityGroup(
                        id=res_id, name=context_[object_type]['name'],
                        tenant_id=context_[object_type]['tenant_id'],
                        description=context_[object_type]['description'])
                    plugin_context.session.add(sg)
                    sg_dict = dict(sg)
                    sg_dict['security_group_rules'] = []
                    with db_api.CONTEXT_WRITER.using(plugin_context):
                        self.mech.sync_from_callback_precommit(
                            plugin_context, operation, res_type, res_id,
                            context_,
                            payload=events.DBEventPayload(plugin_context,
                                                          metadata=sg_dict))
                if operation == odl_const.ODL_DELETE:
                    payload = events.DBEventPayload(
                        context=plugin_context,
                        metadata={
                            'security_group_rule_ids': [SG_RULE_FAKE_ID]},
                        request_body={
                            'security_group_rules': {'id': SG_RULE_FAKE_ID}})
                    with db_api.CONTEXT_WRITER.using(plugin_context):
                        self.mech.sync_from_callback_precommit(
                            plugin_context, operation, res_type, res_id,
                            context_, payload=payload)
            elif (object_type == odl_const.ODL_SG_RULE and
                  operation == odl_const.ODL_DELETE):
                with db_api.CONTEXT_WRITER.using(plugin_context):
                    self.mech.sync_from_callback_precommit(
                        plugin_context, operation, res_type, res_id,
                        context_,
                        payload=events.DBEventPayload(
                            plugin_context,
                            resource_id=SG_FAKE_ID,
                            metadata={'security_group_id': SG_FAKE_ID}))
            else:
                with db_api.CONTEXT_WRITER.using(plugin_context):
                    self.mech.sync_from_callback_precommit(
                        plugin_context, operation, res_type, res_id,
                        context_,
                        payload=events.DBEventPayload(
                            plugin_context,
                            metadata={'security_group_id': [SG_FAKE_ID]}))
        else:
            method = getattr(self.mech, '%s_%s_precommit' % (operation,
                                                             object_type))
            with db_api.CONTEXT_WRITER.using(context):
                method(context)

    def _test_operation_object(self, operation, object_type):
        self._call_operation_object(operation, object_type)

        context = self._get_mock_operation_context(object_type)
        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        self.assertEqual(operation, row['operation'])
        self.assertEqual(object_type, row['object_type'])
        self.assertEqual(context.current['id'], row['object_uuid'])
        self._db_cleanup()

    def _test_thread_processing(self, operation, object_type,
                                expected_calls=1):
        http_requests = {odl_const.ODL_CREATE: 'post',
                         odl_const.ODL_UPDATE: 'put',
                         odl_const.ODL_DELETE: 'delete'}
        status_codes = {odl_const.ODL_CREATE: requests.codes.created,
                        odl_const.ODL_UPDATE: requests.codes.ok,
                        odl_const.ODL_DELETE: requests.codes.no_content}

        http_request = http_requests[operation]
        status_code = status_codes[operation]

        self._call_operation_object(operation, object_type)

        context = self._get_mock_operation_context(object_type)
        url_object_type = utils.neutronify(object_type)
        if operation in [odl_const.ODL_UPDATE, odl_const.ODL_DELETE]:
            if object_type in [odl_const.ODL_SG, odl_const.ODL_SG_RULE]:
                uuid = context[object_type]['id']
            else:
                uuid = context.current['id']
            url = '%s/%ss/%s' % (cfg.CONF.ml2_odl.url, url_object_type, uuid)
        else:
            url = '%s/%ss' % (cfg.CONF.ml2_odl.url, url_object_type)

        if (object_type == odl_const.ODL_SG and
                operation == odl_const.ODL_CREATE):
            context = copy.deepcopy(context)
        if operation in [odl_const.ODL_CREATE, odl_const.ODL_UPDATE]:
            kwargs = {
                'url': url,
                'data': DataMatcher(operation, object_type, context)}
        else:
            kwargs = {'url': url, 'data': None}

        self._test_operation(status_code, expected_calls, http_request,
                             **kwargs)

    def _test_object_type(self, object_type, delete_expected_calls=1):
        # Add and process create request.
        self._test_thread_processing(odl_const.ODL_CREATE, object_type)
        rows = db.get_all_db_rows_by_state(self.db_context,
                                           odl_const.COMPLETED)
        self.assertEqual(1, len(rows))

        # Add and process update request. Adds to database.
        self._test_thread_processing(odl_const.ODL_UPDATE, object_type)
        rows = db.get_all_db_rows_by_state(self.db_context,
                                           odl_const.COMPLETED)
        self.assertEqual(2, len(rows))

        # Add and process update request. Adds to database.
        self._test_thread_processing(odl_const.ODL_DELETE, object_type,
                                     delete_expected_calls)
        rows = db.get_all_db_rows_by_state(self.db_context,
                                           odl_const.COMPLETED)
        self.assertEqual(2 + delete_expected_calls, len(rows))

    def _test_object_type_pending_network(self, object_type):
        # Create a network (creates db row in pending state).
        self._call_operation_object(odl_const.ODL_CREATE,
                                    odl_const.ODL_NETWORK)

        # Create object_type database row and process. This results in both
        # the object_type and network rows being processed.
        self._test_thread_processing(odl_const.ODL_CREATE, object_type,
                                     expected_calls=2)

        # Verify both rows are now marked as completed.
        rows = db.get_all_db_rows_by_state(self.db_context,
                                           odl_const.COMPLETED)
        self.assertEqual(2, len(rows))

    def _test_object_type_processing_network(self, object_type):
        self._test_object_operation_pending_another_object_operation(
            object_type, odl_const.ODL_CREATE, odl_const.ODL_NETWORK,
            odl_const.ODL_CREATE)

    def _test_object_operation_pending_object_operation(
            self, object_type, operation, pending_operation):
        self._test_object_operation_pending_another_object_operation(
            object_type, operation, object_type, pending_operation)

    def _test_object_operation_pending_another_object_operation(
            self, object_type, operation, pending_type, pending_operation):
        # Create the object_type (creates db row in pending state).
        self._call_operation_object(pending_operation,
                                    pending_type)

        # Get pending row and mark as processing so that
        # this row will not be processed by journal thread.
        row = db.get_all_db_rows_by_state(self.db_context, odl_const.PENDING)
        db.update_db_row_state(self.db_context, row[0], odl_const.PROCESSING)

        # Create the object_type database row and process.
        # Verify that object request is not processed because the
        # dependent object operation has not been marked as 'completed'.
        self._test_thread_processing(operation,
                                     object_type,
                                     expected_calls=0)

        # Verify that all rows are still in the database.
        rows = db.get_all_db_rows_by_state(self.db_context,
                                           odl_const.PROCESSING)
        self.assertEqual(1, len(rows))
        rows = db.get_all_db_rows_by_state(self.db_context, odl_const.PENDING)
        self.assertEqual(1, len(rows))

    def _test_parent_delete_pending_child_delete(self, parent, child):
        self._test_object_operation_pending_another_object_operation(
            parent, odl_const.ODL_DELETE, child, odl_const.ODL_DELETE)

    def _test_cleanup_processing_rows(self, last_retried, expected_state):
        # Create a dummy network (creates db row in pending state).
        self._call_operation_object(odl_const.ODL_CREATE,
                                    odl_const.ODL_NETWORK)

        # Get pending row and mark as processing and update
        # the last_retried time
        row = db.get_all_db_rows_by_state(self.db_context,
                                          odl_const.PENDING)[0]
        row.last_retried = last_retried
        db.update_db_row_state(self.db_context, row, odl_const.PROCESSING)

        # Test if the cleanup marks this in the desired state
        # based on the last_retried timestamp
        cleanup.cleanup_processing_rows(self.db_context)

        # Verify that the Db row is in the desired state
        rows = db.get_all_db_rows_by_state(self.db_context, expected_state)
        self.assertEqual(1, len(rows))

    def test_driver(self):
        for operation in [odl_const.ODL_CREATE, odl_const.ODL_UPDATE,
                          odl_const.ODL_DELETE]:
            for object_type in [odl_const.ODL_NETWORK, odl_const.ODL_SUBNET,
                                odl_const.ODL_PORT]:
                self._test_operation_object(operation, object_type)

    def test_port_precommit_no_tenant(self):
        context = self._get_mock_operation_context(odl_const.ODL_PORT)
        context.current['tenant_id'] = ''

        method = getattr(self.mech, 'create_port_precommit')
        method(context)
        self.db_context.session.flush()

        # Verify that the Db row has a tenant
        rows = db.get_all_db_rows_by_state(self.db_context, odl_const.PENDING)
        self.assertEqual(1, len(rows))
        _network = self._get_mock_network_operation_context().current
        self.assertEqual(_network['tenant_id'], rows[0]['data']['tenant_id'])

    def test_network(self):
        self._test_object_type(odl_const.ODL_NETWORK)

    def test_network_update_pending_network_create(self):
        self._test_object_operation_pending_object_operation(
            odl_const.ODL_NETWORK, odl_const.ODL_UPDATE, odl_const.ODL_CREATE)

    def test_network_delete_pending_network_create(self):
        self._test_object_operation_pending_object_operation(
            odl_const.ODL_NETWORK, odl_const.ODL_DELETE, odl_const.ODL_CREATE)

    def test_network_delete_pending_network_update(self):
        self._test_object_operation_pending_object_operation(
            odl_const.ODL_NETWORK, odl_const.ODL_DELETE, odl_const.ODL_UPDATE)

    def test_network_delete_pending_subnet_delete(self):
        self._test_parent_delete_pending_child_delete(
            odl_const.ODL_NETWORK, odl_const.ODL_SUBNET)

    def test_network_delete_pending_port_delete(self):
        self._test_parent_delete_pending_child_delete(
            odl_const.ODL_NETWORK, odl_const.ODL_PORT)

    def test_subnet(self):
        self._test_object_type(odl_const.ODL_SUBNET)

    def test_subnet_update_pending_subnet_create(self):
        self._test_object_operation_pending_object_operation(
            odl_const.ODL_SUBNET, odl_const.ODL_UPDATE, odl_const.ODL_CREATE)

    def test_subnet_delete_pending_subnet_create(self):
        self._test_object_operation_pending_object_operation(
            odl_const.ODL_SUBNET, odl_const.ODL_DELETE, odl_const.ODL_CREATE)

    def test_subnet_delete_pending_subnet_update(self):
        self._test_object_operation_pending_object_operation(
            odl_const.ODL_SUBNET, odl_const.ODL_DELETE, odl_const.ODL_UPDATE)

    def test_subnet_pending_network(self):
        self._test_object_type_pending_network(odl_const.ODL_SUBNET)

    def test_subnet_processing_network(self):
        self._test_object_type_processing_network(odl_const.ODL_SUBNET)

    def test_subnet_delete_pending_port_delete(self):
        self._test_parent_delete_pending_child_delete(
            odl_const.ODL_SUBNET, odl_const.ODL_PORT)

    def test_port(self):
        self._test_object_type(odl_const.ODL_PORT)

    def test_port_update_pending_port_create(self):
        self._test_object_operation_pending_object_operation(
            odl_const.ODL_PORT, odl_const.ODL_UPDATE, odl_const.ODL_CREATE)

    def test_port_delete_pending_port_create(self):
        self._test_object_operation_pending_object_operation(
            odl_const.ODL_PORT, odl_const.ODL_DELETE, odl_const.ODL_CREATE)

    def test_port_delete_pending_port_update(self):
        self._test_object_operation_pending_object_operation(
            odl_const.ODL_PORT, odl_const.ODL_DELETE, odl_const.ODL_UPDATE)

    def test_port_pending_network(self):
        self._test_object_type_pending_network(odl_const.ODL_PORT)

    def test_port_processing_network(self):
        self._test_object_type_processing_network(odl_const.ODL_PORT)

    def test_cleanup_processing_rows_time_not_expired(self):
        self._test_cleanup_processing_rows(datetime.datetime.utcnow(),
                                           odl_const.PROCESSING)

    def test_cleanup_processing_rows_time_expired(self):
        old_time = datetime.datetime.utcnow() - datetime.timedelta(hours=24)
        self._test_cleanup_processing_rows(old_time, odl_const.PENDING)

    def test_thread_call(self):
        """Verify that the sync thread method is called."""

        with mock.patch.object(
                journal.OpenDaylightJournalThread,
                'start') as mock_sync_thread:
            self.mech = mech_driver_v2.OpenDaylightMechanismDriver()
            self.mech.initialize()

            # Create any object that would spin up the sync thread via the
            # decorator call_thread_on_end() used by all the event handlers.
            self._call_operation_object(odl_const.ODL_CREATE,
                                        odl_const.ODL_NETWORK)

            # Verify that the thread call was made.
            mock_sync_thread.assert_called()

    def test_sg(self):
        self._test_object_type(odl_const.ODL_SG, 2)

    def test_sg_rule(self):
        self._test_object_type(odl_const.ODL_SG_RULE)

    def test_sg_delete(self):
        with mock.patch.object(journal, 'record') as record:
            context = self._get_mock_operation_context(odl_const.ODL_SG)
            res_id = context[odl_const.ODL_SG]['id']
            rule = mock.Mock()
            rule.id = SG_RULE_FAKE_ID
            rule.security_group_id = SG_FAKE_ID
            sg = mock.Mock()
            sg.id = SG_FAKE_ID
            sg.security_group_rules = [rule]
            payload = events.DBEventPayload(
                self.db_context,
                metadata={'security_group_rule_ids': [SG_RULE_FAKE_ID]}
            )
            with db_api.CONTEXT_WRITER.using(self.db_context):
                self.mech.sync_from_callback_precommit(
                    self.db_context, odl_const.ODL_DELETE,
                    callback._RESOURCE_MAPPING[odl_const.ODL_SG],
                    res_id, context, payload=payload)
            record.assert_has_calls(
                [mock.call(mock.ANY, 'security_group_rule',
                           SG_RULE_FAKE_ID, 'delete', [SG_FAKE_ID]),
                 mock.call(mock.ANY, 'security_group', SG_FAKE_ID,
                           'delete',
                           {'description': 'test-description',
                            'project_id': 'test-tenant',
                            'security_group_rules': [],
                            'tenant_id': 'test-tenant',
                            'id': SG_FAKE_ID, 'name': 'test_sg'})])

    def test_sg_rule_delete(self):
        with mock.patch.object(journal, 'record') as record:
            context = self._get_mock_operation_context(odl_const.ODL_SG_RULE)
            res_id = context[odl_const.ODL_SG_RULE]['id']
            rule = mock.Mock()
            rule.id = SG_RULE_FAKE_ID
            rule.security_group_id = SG_FAKE_ID
            with db_api.CONTEXT_WRITER.using(self.db_context):
                payload = events.DBEventPayload(
                    self.db_context,
                    metadata={'security_group_id': SG_FAKE_ID})
                self.mech.sync_from_callback_precommit(
                    self.db_context, odl_const.ODL_DELETE,
                    callback._RESOURCE_MAPPING[odl_const.ODL_SG_RULE],
                    res_id, context, payload=payload)
            record.assert_has_calls(
                [mock.call(mock.ANY, 'security_group_rule',
                           SG_RULE_FAKE_ID, 'delete', [SG_FAKE_ID])])

    def test_subnet_allocation_pools(self):
        context = self._get_mock_operation_context(odl_const.ODL_SUBNET)
        alloc_pool = context.current['allocation_pools']
        self._call_operation_object(odl_const.ODL_UPDATE,
                                    odl_const.ODL_SUBNET)
        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        self.assertEqual(alloc_pool, row.data['allocation_pools'])

    def test_sync_multiple_updates(self):
        # add 2 updates
        for i in range(2):
            self._call_operation_object(odl_const.ODL_UPDATE,
                                        odl_const.ODL_NETWORK)

        # get the last update row
        rows = db.get_all_db_rows(self.db_context)
        rows.sort(key=operator.attrgetter("seqnum"))
        first_row = rows[0]

        # change the state to processing
        db.update_db_row_state(self.db_context, first_row,
                               odl_const.PROCESSING)

        # create 1 more operation to trigger the sync thread
        # verify that there are no calls to ODL controller, because the
        # first row was processing (exit_after_run = true)
        self._test_thread_processing(odl_const.ODL_UPDATE,
                                     odl_const.ODL_NETWORK, expected_calls=0)

        # validate that all the pending rows stays in 'pending' state
        # first row should be 'processing' because it was not processed
        processing = db.get_all_db_rows_by_state(self.db_context, 'processing')
        self.assertEqual(1, len(processing))
        rows = db.get_all_db_rows_by_state(self.db_context, 'pending')
        self.assertEqual(2, len(rows))

    def test_update_port_filter(self):
        """Validate the filter code on update port operation"""
        expected_items = ['fixed_ips', 'security_groups', 'device_id',
                          'security_groups', 'admin_state_up']
        subnet = self._get_mock_operation_context(odl_const.ODL_SUBNET).current
        port = self._get_mock_operation_context(odl_const.ODL_PORT).current
        port['fixed_ips'] = [{'subnet_id': subnet['id'],
                              'ip_address': '10.0.0.10'}]
        port['mac_address'] = port['mac_address'].upper()

        orig_port = copy.deepcopy(port)

        with mock.patch.object(segments_db, 'get_network_segments'):
            filters.filter_for_odl(odl_const.ODL_PORT,
                                   odl_const.ODL_UPDATE, port)
            for key, value in orig_port.items():
                if key in expected_items:
                    self.assertEqual(port[key], value)


class _OpenDaylightDriverVlanTransparencyBase(_OpenDaylightMechanismBase):
    def _driver_context(self, network):
        return mock.MagicMock(current=network)


class TestOpenDaylightDriverVlanTransparencyNetwork(
        _OpenDaylightDriverVlanTransparencyBase):
    def _test_network_type(self, expected, network_type):
        context = self._driver_context({providernet.NETWORK_TYPE:
                                        network_type})
        self.assertEqual(expected,
                         self.mech.check_vlan_transparency(context))

    def test_none_network_type(self):
        context = self._driver_context({})
        self.assertTrue(self.mech.check_vlan_transparency(context))

    def test_vlan_transparency(self):
        for network_type in [n_constants.TYPE_VXLAN]:
            self._test_network_type(True, network_type)
        for network_type in [n_constants.TYPE_FLAT, n_constants.TYPE_GENEVE,
                             n_constants.TYPE_GRE, n_constants.TYPE_LOCAL,
                             n_constants.TYPE_VLAN]:
            self._test_network_type(False, network_type)


class TestOpenDaylightDriverVlanTransparency(
        _OpenDaylightDriverVlanTransparencyBase):
    scenarios = [
        ("vxlan_vxlan",
         {'expected': True,
          'network_types': [n_constants.TYPE_VXLAN, n_constants.TYPE_VXLAN]}),
        ("gre_vxlan",
         {'expected': False,
          'network_types': [n_constants.TYPE_GRE, n_constants.TYPE_VXLAN]}),
        ("vxlan_vlan",
         {'expected': False,
          'network_types': [n_constants.TYPE_VXLAN, n_constants.TYPE_VLAN]}),
        ("vxlan_flat",
         {'expected': False,
          'network_types': [n_constants.TYPE_VXLAN, n_constants.TYPE_FLAT]}),
        ("vlan_vlan",
         {'expected': False,
          'network_types': [n_constants.TYPE_VLAN, n_constants.TYPE_VLAN]}),
    ]

    def test_network_segments(self):
        segments = [{providernet.NETWORK_TYPE: type_}
                    for type_ in self.network_types]
        context = self._driver_context({mpnet_apidef.SEGMENTS: segments})
        self.assertEqual(self.expected,
                         self.mech.check_vlan_transparency(context))



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fnmatch
import re

from oslo_serialization import jsonutils


def json(obj):
    return MatchJson(obj)


class MatchJson(object):

    def __init__(self, obj):
        self._obj = obj

    def __eq__(self, json_text):
        return self._obj == jsonutils.loads(json_text)

    def __repr__(self):
        return "MatchJson({})".format(repr(self._obj))


def wildcard(text):
    return MatchWildcard(text)


class MatchWildcard(object):

    def __init__(self, obj):
        self._text = text = str(obj)
        self._reg = re.compile(fnmatch.translate(text))

    def __eq__(self, obj):
        return self._reg.match(str(obj))

    def __repr__(self):
        return "MatchWildcard({})".format(self._text)



# Copyright (c) 2016 Brocade Communication Systems
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from networking_sfc.extensions import flowclassifier as fc_const
from networking_sfc.services.flowclassifier.drivers import base as fc_driver

from networking_odl.common import constants as odl_const
from networking_odl.common import postcommit
from networking_odl.journal import full_sync
from networking_odl.journal import journal

LOG = logging.getLogger(__name__)

SFC_FC_RESOURCES = {
    odl_const.ODL_SFC_FLOW_CLASSIFIER:
        odl_const.NETWORKING_SFC_FLOW_CLASSIFIERS,
}


@postcommit.add_postcommit('flow_classifier')
class OpenDaylightSFCFlowClassifierDriverV2(
        fc_driver.FlowClassifierDriverBase):

    """OpenDaylight SFC Flow Classifier Driver (Version 2) for networking-sfc.

    This Driver pass through SFC Flow Classifier API calls to
    OpenDaylight Neutron Northbound Project by using the REST
    API's exposed by the project.
    """

    def initialize(self):
        LOG.debug("Initializing OpenDaylight Networking "
                  "SFC Flow Classifier driver Version 2")
        self.journal = journal.OpenDaylightJournalThread()
        full_sync.register(fc_const.FLOW_CLASSIFIER_EXT, SFC_FC_RESOURCES)

    @staticmethod
    def _record_in_journal(context, object_type, operation, data=None):
        if data is None:
            data = context.current
        journal.record(context._plugin_context, object_type,
                       context.current['id'], operation, data)

    @log_helpers.log_method_call
    def create_flow_classifier_precommit(self, context):
        OpenDaylightSFCFlowClassifierDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_FLOW_CLASSIFIER, odl_const.ODL_CREATE)

    @log_helpers.log_method_call
    def update_flow_classifier_precommit(self, context):
        OpenDaylightSFCFlowClassifierDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_FLOW_CLASSIFIER, odl_const.ODL_UPDATE)

    @log_helpers.log_method_call
    def delete_flow_classifier_precommit(self, context):
        OpenDaylightSFCFlowClassifierDriverV2._record_in_journal(
            context, odl_const.ODL_SFC_FLOW_CLASSIFIER, odl_const.ODL_DELETE,
            data=[])

    # Need to implement these methods, else driver loading fails with error
    # complaining about no abstract method implementation present.
    @log_helpers.log_method_call
    def create_flow_classifier(self, context):
        super(OpenDaylightSFCFlowClassifierDriverV2,
              self).create_flow_classifier(context)

    @log_helpers.log_method_call
    def update_flow_classifier(self, context):
        super(OpenDaylightSFCFlowClassifierDriverV2,
              self).update_flow_classifier(context)

    @log_helpers.log_method_call
    def delete_flow_classifier(self, context):
        super(OpenDaylightSFCFlowClassifierDriverV2,
              self).delete_flow_classifier(context)



# Copyright 2016 Intel Corporation.
# Copyright 2016 Isaku Yamahata <isaku.yamahata at intel com>
#                               <isaku.yamahata at gmail com>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from sqlalchemy import sql
from sqlalchemy.sql import schema

from neutron.db.migration.alembic_migrations import external
from neutron.db.migration import cli as migration
from neutron.tests.functional.db import test_migrations
from neutron.tests.unit import testlib_api

from networking_odl.db import head

FWAAS_TABLES = (
    'cisco_firewall_associations',
    'firewall_group_port_associations_v2',
    'firewall_groups_v2',
    'firewall_policies_v2',
    'firewall_policy_rule_associations_v2',
    'firewall_router_associations',
    'firewall_rules_v2',
)

L2GW_TABLES = (
    'l2gatewayconnections',
    'l2gatewaydevices',
    'l2gatewayinterfaces',
    'l2gateways',
    'l2gw_alembic_version',
    'logical_switches',
    'pending_ucast_macs_remotes',
    'physical_locators',
    'physical_ports',
    'physical_switches',
    'ucast_macs_locals',
    'ucast_macs_remotes',
    'vlan_bindings',
)

BGPVPN_TABLES = (
    'bgpvpns',
    'bgpvpn_network_associations',
    'bgpvpn_router_associations',
    'ml2_route_target_allocations',
    'sfc_bagpipe_ppg_rtnn_associations',
    'sfc_bagpipe_chain_hops',
)

# Tables from other repos that we depend on but do not manage.
IGNORED_TABLES_MATCH = set(
    FWAAS_TABLES + L2GW_TABLES + BGPVPN_TABLES
)

# EXTERNAL_TABLES should contain all names of tables that are not related to
# current repo.
EXTERNAL_TABLES = set(external.TABLES)

VERSION_TABLE = 'odl_alembic_version'


class _TestModelsMigrationsODL(test_migrations._TestModelsMigrations):
    def db_sync(self, engine):
        self.cfg.config(connection=engine.url, group='database')
        for conf in migration.get_alembic_configs():
            self.alembic_config = conf
            self.alembic_config.neutron_config = cfg.CONF
            migration.do_alembic_command(conf, 'upgrade', 'heads')

    def get_metadata(self):
        return head.get_metadata()

    def include_object(self, object_, name, type_, reflected, compare_to):
        if type_ == 'table' and (name.startswith('alembic') or
                                 name == VERSION_TABLE or
                                 name in EXTERNAL_TABLES or
                                 name in IGNORED_TABLES_MATCH):
            return False
        if type_ == 'index' and reflected and name.startswith("idx_autoinc_"):
            return False
        return True

    def _filter_mysql_server_func_now(self, diff_elem):
        # TODO(yamahata): remove this bug work around once it's fixed
        # example:
        # when the column has server_default=sa.func.now(), the diff
        # includes the followings diff
        # [ ('modify_default',
        #     None,
        #    'opendaylightjournal',
        #    'created_at',
        #   {'existing_nullable': True,
        #    'existing_type': DATETIME()},
        # DefaultClause(<sqlalchemy.sql.elements.TextClause object
        #               at 0x7f652188ce50>, for_update=False),
        # DefaultClause(<sqlalchemy.sql.functions.now at 0x7f6522411050; now>,
        #               for_update=False))]
        # another example
        # [ ('modify_default',
        #     None,
        #    'opendaylightjournal',
        #    'created_at',
        #   {'existing_nullable': True,
        #    'existing_type': DATETIME()},
        #     None,
        #   DefaultClause(<sqlalchemy.sql.functions.now at 0x7ff3b3517410;
        #                  now>,
        #                 for_update=False))]

        meta_def = diff_elem[0][5]
        rendered_meta_def = diff_elem[0][6]
        if (diff_elem[0][0] == 'modify_default' and
                diff_elem[0][2] in ('opendaylightjournal',
                                    'opendaylight_periodic_task') and
                isinstance(meta_def, schema.DefaultClause) and
                isinstance(meta_def.arg, sql.elements.TextClause) and
                meta_def.reflected and
                meta_def.arg.text == 'CURRENT_TIMESTAMP' and
                isinstance(rendered_meta_def, schema.DefaultClause) and
                isinstance(rendered_meta_def.arg, sql.functions.now) and not
                rendered_meta_def.reflected and
                meta_def.for_update == rendered_meta_def.for_update):
            return False

        return True

    def filter_metadata_diff(self, diff):
        return list(filter(self._filter_mysql_server_func_now, diff))


class TestModelsMigrationsMysql(testlib_api.MySQLTestCaseMixin,
                                _TestModelsMigrationsODL,
                                testlib_api.SqlTestCaseLight):
    pass


class TestModelsMigrationsPostgresql(testlib_api.PostgreSQLTestCaseMixin,
                                     _TestModelsMigrationsODL,
                                     testlib_api.SqlTestCaseLight):
    pass



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron.plugins.ml2 import driver_context as ctx
from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_constants
from neutron_lib.plugins.ml2 import api

from networking_odl.ml2 import legacy_port_binding
from networking_odl.tests import base


class TestLegacyPortBindingManager(base.DietTestCase):
    # valid  and invalid segments
    valid_segment = {
        api.ID: 'API_ID',
        api.NETWORK_TYPE: n_constants.TYPE_LOCAL,
        api.SEGMENTATION_ID: 'API_SEGMENTATION_ID',
        api.PHYSICAL_NETWORK: 'API_PHYSICAL_NETWORK'}

    invalid_segment = {
        api.ID: 'API_ID',
        api.NETWORK_TYPE: n_constants.TYPE_NONE,
        api.SEGMENTATION_ID: 'API_SEGMENTATION_ID',
        api.PHYSICAL_NETWORK: 'API_PHYSICAL_NETWORK'}

    def test_check_segment(self):
        """Validate the _check_segment method."""

        all_network_types = [n_constants.TYPE_FLAT, n_constants.TYPE_GRE,
                             n_constants.TYPE_LOCAL, n_constants.TYPE_VXLAN,
                             n_constants.TYPE_VLAN, n_constants.TYPE_NONE]

        mgr = legacy_port_binding.LegacyPortBindingManager()

        valid_types = {
            network_type
            for network_type in all_network_types
            if mgr._check_segment({api.NETWORK_TYPE: network_type})}

        self.assertEqual({
            n_constants.TYPE_FLAT, n_constants.TYPE_LOCAL,
            n_constants.TYPE_GRE, n_constants.TYPE_VXLAN,
            n_constants.TYPE_VLAN}, valid_types)

    def test_bind_port(self):

        network = mock.MagicMock(spec=api.NetworkContext)

        port_context = mock.MagicMock(
            spec=ctx.PortContext, current={'id': 'CURRENT_CONTEXT_ID'},
            segments_to_bind=[self.valid_segment, self.invalid_segment],
            network=network)

        mgr = legacy_port_binding.LegacyPortBindingManager()
        vif_type = mgr._get_vif_type(port_context)

        mgr.bind_port(port_context)

        port_context.set_binding.assert_called_once_with(
            self.valid_segment[api.ID], vif_type,
            mgr.vif_details, status=n_constants.PORT_STATUS_ACTIVE)

    def test_bind_port_unsupported_vnic_type(self):
        network = mock.MagicMock(spec=api.NetworkContext)
        port_context = mock.MagicMock(
            spec=ctx.PortContext,
            current={'id': 'CURRENT_CONTEXT_ID',
                     portbindings.VNIC_TYPE: portbindings.VNIC_DIRECT},
            segments_to_bind=[self.valid_segment, self.invalid_segment],
            network=network)

        mgr = legacy_port_binding.LegacyPortBindingManager()
        mgr.bind_port(port_context)
        port_context.set_binding.assert_not_called()



#
# Copyright (C) 2017 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from unittest import mock

from networking_odl.db import db
from networking_odl.journal import cleanup
from networking_odl.tests.unit import test_base_db


class JournalCleanupTestCase(test_base_db.ODLBaseDbTestCase):
    def test_delete_completed_rows_retries_exceptions(self):
        self.cfg.config(completed_rows_retention=1, group='ml2_odl')
        with mock.patch.object(db, 'delete_rows_by_state_and_time') as m:
            self._test_retry_exceptions(
                cleanup.delete_completed_rows, m)

    def test_cleanup_processsing_rows_retries_exceptions(self):
        with mock.patch.object(db, 'reset_processing_rows') as m:
            self._test_retry_exceptions(
                cleanup.cleanup_processing_rows, m)

    @mock.patch.object(db, 'delete_rows_by_state_and_time')
    def _test_delete_completed_rows(self, retention, expected_call, mock_db):
        self.cfg.config(completed_rows_retention=retention, group='ml2_odl')
        cleanup.delete_completed_rows(self.db_context)
        self.assertEqual(expected_call, mock_db.called)

    def test_delete_completed_rows_with_retention(self):
        self._test_delete_completed_rows(1, True)

    def test_delete_completed_rows_zero_retention(self):
        self._test_delete_completed_rows(0, False)

    def test_delete_completed_rows_indefinite_retention(self):
        self._test_delete_completed_rows(-1, False)



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron_lib import constants as n_const
from neutron_lib.plugins import directory
from oslo_config import fixture as config_fixture

from networking_odl.common import constants as odl_const
from networking_odl.tests.functional import base


class TestOdlDhcpDriver(base.V2DriverAdjustment, base.OdlTestsBase,
                        test_plugin.Ml2PluginV2TestCase):
    _mechanism_drivers = ['opendaylight_v2']

    def setUp(self):
        self.cfg = self.useFixture(config_fixture.Config())
        self.cfg.config(enable_dhcp_service=True, group='ml2_odl')
        super(TestOdlDhcpDriver, self).setUp()

    def get_port_data(self, network, subnet):
        plugin = self.get_plugin()
        device_id = odl_const.ODL_DEVICE_ID_START + \
            '-' + subnet[odl_const.ODL_SUBNET]['id']
        filters = {
            'network_id': [network[odl_const.ODL_NETWORK]['id']],
            'device_id': [device_id],
            'device_owner': [n_const.DEVICE_OWNER_DHCP]
        }
        ports = plugin.get_ports(self.context, filters=filters)
        if ports:
            port = ports[0]
            return {odl_const.ODL_PORT: {'id': port['id']}}

    def get_plugin(self):
        return directory.get_plugin()

    def test_subnet_create(self):
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                self.get_odl_resource(odl_const.ODL_SUBNET, subnet)
                port = self.get_port_data(network, subnet)
                self.assert_resource_created(odl_const.ODL_PORT, port)

    def test_subnet_update_from_disable_to_enable(self):
        with self.network() as network:
            with self.subnet(network=network, enable_dhcp=False) as subnet:
                self.get_odl_resource(odl_const.ODL_SUBNET, subnet)
                plugin = self.get_plugin()
                port = self.get_port_data(network, subnet)
                self.assertIsNone(port)
                subnet[odl_const.ODL_SUBNET]['enable_dhcp'] = True
                plugin.update_subnet(
                    self.context, subnet[odl_const.ODL_SUBNET]['id'], subnet)
                self.get_odl_resource(odl_const.ODL_SUBNET, subnet)
                port = self.get_port_data(network, subnet)
                self.assert_resource_created(odl_const.ODL_PORT, port)

    def test_subnet_update_from_enable_to_disable(self):
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                self.get_odl_resource(odl_const.ODL_SUBNET, subnet)
                plugin = self.get_plugin()
                port = self.get_port_data(network, subnet)
                self.assert_resource_created(odl_const.ODL_PORT, port)

                subnet[odl_const.ODL_SUBNET]['enable_dhcp'] = False
                plugin.update_subnet(
                    self.context, subnet[odl_const.ODL_SUBNET]['id'], subnet)
                resource = self.get_odl_resource(odl_const.ODL_PORT, port)
                self.assertIsNone(resource)

    def test_subnet_delete(self):
        with self.network() as network:
            with self.subnet(network=network) as subnet:
                self.get_odl_resource(odl_const.ODL_SUBNET, subnet)
                plugin = self.get_plugin()
                port = self.get_port_data(network, subnet)
                self.assert_resource_created(odl_const.ODL_PORT, port)
                plugin.delete_subnet(
                    self.context, subnet[odl_const.ODL_SUBNET]['id'])
                resource = self.get_odl_resource(odl_const.ODL_PORT, port)
                self.assertIsNone(resource)



#
# Copyright (C) 2016 Red Hat, Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#

from neutron_lib.db import api as db_api
from neutron_lib import exceptions as nexc
from neutron_lib.plugins import directory
from oslo_log import log as logging

from networking_odl.common import client
from networking_odl.common import constants as odl_const
from networking_odl.common import exceptions
from networking_odl.db import db
from networking_odl.journal import base_driver
from networking_odl.journal import full_sync
from networking_odl.journal import journal

_CLIENT = client.OpenDaylightRestClientGlobal()

LOG = logging.getLogger(__name__)


@db_api.retry_if_session_inactive()
def journal_recovery(context):
    for row in db.get_all_db_rows_by_state(context, odl_const.FAILED):
        LOG.debug("Attempting recovery of journal entry %s.", row)
        try:
            odl_resource = _CLIENT.get_client().get_resource(
                row.object_type,
                row.object_uuid)
        except exceptions.UnsupportedResourceType:
            LOG.warning('Unsupported resource %s', row.object_type)
        except Exception:
            LOG.exception("Failure while recovering journal entry %s.", row)
        else:
            with db_api.CONTEXT_WRITER.savepoint.using(context):
                if odl_resource is not None:
                    _handle_existing_resource(context, row)
                else:
                    _handle_non_existing_resource(context, row)


def get_latest_resource(context, row):
    try:
        driver = base_driver.get_driver(row.object_type)
    except exceptions.ResourceNotRegistered:
        raise exceptions.UnsupportedResourceType(
            resource=row.object_type) from None

    return driver.get_resource_for_recovery(context, row)


# TODO(rajivk): Remove this method once recovery is fully supported
def _get_latest_resource(context, row):
    object_type = row.object_type

    for plugin_alias, resources in full_sync.ALL_RESOURCES.items():
        if object_type in resources:
            plugin = directory.get_plugin(plugin_alias)
            break
    else:
        raise exceptions.UnsupportedResourceType(resource=object_type)

    obj_getter = getattr(plugin, 'get_{}'.format(object_type))
    return obj_getter(context, row.object_uuid)


def _sync_resource_to_odl(context, row, operation_type, exists_on_odl):
    resource = None
    try:
        resource = _get_latest_resource(context, row)
    except nexc.NotFound:
        if exists_on_odl:
            journal.record(context, row.object_type,
                           row.object_uuid, odl_const.ODL_DELETE, [])
    else:
        journal.record(context, row.object_type, row.object_uuid,
                       operation_type, resource)

    journal.entry_complete(context, row)


def _handle_existing_resource(context, row):
    if row.operation == odl_const.ODL_CREATE:
        journal.entry_complete(context, row)
    elif row.operation == odl_const.ODL_DELETE:
        db.update_db_row_state(context, row, odl_const.PENDING)
    else:
        _sync_resource_to_odl(context, row, odl_const.ODL_UPDATE, True)


def _handle_non_existing_resource(context, row):
    if row.operation == odl_const.ODL_DELETE:
        journal.entry_complete(context, row)
    else:
        _sync_resource_to_odl(context, row, odl_const.ODL_CREATE, False)
        # TODO(mkolesni): Handle missing parent resources somehow.



# Copyright (c) 2017 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_config import fixture as config_fixture
from oslo_serialization import jsonutils
from requests import exceptions

from networking_odl.common.client import OpenDaylightRestClient
from networking_odl.common import odl_features
from networking_odl.tests import base


class TestOdlFeatures(base.DietTestCase):
    """Basic tests for odl_features"""

    feature_json = """{"features": {"feature":
                            [{"service-provider-feature":
                            "neutron-extensions:operational-port-status"},
                            {"service-provider-feature":
                            "neutron-extensions:feature-with-config",
                            "configuration": "steal-your-face"}]}}"""

    feature_list = 'thing1, thing2'

    def setUp(self):
        self.features_fixture = base.OpenDaylightFeaturesFixture()
        self.useFixture(self.features_fixture)
        self.cfg = self.useFixture(config_fixture.Config())
        super(TestOdlFeatures, self).setUp()
        self.addCleanup(odl_features.deinit)

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_fetch_exception(self, mocked_client):
        mocked_client.side_effect = exceptions.ConnectionError()
        self.assertIsNone(odl_features._fetch_features())

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_fetch_404(self, mocked_client):
        mocked_client.return_value = mock.MagicMock(status_code=404)
        self.assertNotEqual(id(odl_features._fetch_features()),
                            id(odl_features.EMPTY_FEATURES))

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_fetch_400(self, mocked_client):
        mocked_client.return_value = mock.MagicMock(status_code=400)
        self.assertNotEqual(id(odl_features._fetch_features()),
                            id(odl_features.EMPTY_FEATURES))

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_fetch_500(self, mocked_client):
        mocked_client.return_value = mock.MagicMock(status_code=500)
        self.assertIsNone(odl_features._fetch_features())

    @mock.patch.object(OpenDaylightRestClient, 'request')
    def test_init(self, mocked_client):
        self.cfg.config(odl_features=None, group='ml2_odl')
        self.cfg.config(odl_features_json=None, group='ml2_odl')
        response = mock.MagicMock()
        response.status_code = 200
        response.json = mock.MagicMock(
            return_value=jsonutils.loads(self.feature_json))
        mocked_client.return_value = response

        self._assert_odl_feature_config({
            odl_features.OPERATIONAL_PORT_STATUS: '',
            'feature-with-config': 'steal-your-face',
        })

    def _assert_odl_feature_config(self, features):
        odl_features.init()
        for k, v in features.items():
            self.assertTrue(odl_features.has(k))
            self.assertEqual(odl_features.get_config(k), v)

    def test_init_from_config_json(self):
        self.cfg.config(odl_features_json=self.feature_json, group='ml2_odl')

        self._assert_odl_feature_config({
            odl_features.OPERATIONAL_PORT_STATUS: '',
            'feature-with-config': 'steal-your-face',
        })

    @mock.patch.object(odl_features, '_fetch_features')
    def test_init_without_config_calls__fetch_features(self, mock_fetch):
        self.cfg.config(odl_features_json=None, group='ml2_odl')
        self.cfg.config(odl_features=None, group='ml2_odl')
        odl_features.init()
        mock_fetch.assert_called_once()

    @mock.patch.object(odl_features, '_fetch_features')
    def test_init_from_config_list(self, mock_fetch):
        self.cfg.config(odl_features_json=None, group='ml2_odl')
        self.cfg.config(odl_features=self.feature_list, group='ml2_odl')
        odl_features.init()
        self.assertTrue(odl_features.has('thing1'))
        self.assertTrue(odl_features.has('thing2'))
        mock_fetch.assert_not_called()

    @mock.patch.object(odl_features, '_fetch_features')
    def test_init_from_json_overrides_list(self, mock_fetch):
        self.cfg.config(odl_features=self.feature_list, group='ml2_odl')
        self.cfg.config(odl_features_json=self.feature_json, group='ml2_odl')
        odl_features.init()
        self.assertFalse(odl_features.has('thing1'))
        self.assertTrue(odl_features.has('operational-port-status'))
        mock_fetch.assert_not_called()

    @mock.patch.object(odl_features, '_fetch_features')
    def test_init_with_config_does_not_call__fetch_features(self, mock_fetch):
        self.cfg.config(odl_features_json=self.feature_json, group='ml2_odl')
        odl_features.init()
        mock_fetch.assert_not_called()

    def test_init_from_config_malformed_json_raises_exception(self):
        malformed_json = ")]}'" + self.feature_json
        self.cfg.config(odl_features_json=malformed_json, group='ml2_odl')
        self.assertRaises(ValueError, odl_features.init)

    def test_init_from_config_json_not_in_odl_format_raises_exception(self):
        self.cfg.config(odl_features_json='{}', group='ml2_odl')
        self.assertRaises(KeyError, odl_features.init)


class TestOdlFeaturesNoFixture(base.DietTestCase):
    """Basic tests for odl_features that don't call init()"""

    def setUp(self):
        super(TestOdlFeaturesNoFixture, self).setUp()
        self.addCleanup(odl_features.deinit)

    def test_feature_configs_does_not_mutate_default_features(self):
        odl_features.deinit()
        self.assertEqual(odl_features.EMPTY_FEATURES,
                         odl_features.feature_configs)
        odl_features.feature_configs['test'] = True
        self.assertNotEqual(odl_features.EMPTY_FEATURES,
                            odl_features.feature_configs)

    def test_deinit_does_not_mutate_default_features(self):
        # we call it before initing anything, to force the reassignment
        # of the global features variable.
        odl_features.deinit()
        odl_features.feature_configs['test'] = True
        self.assertNotEqual(odl_features.EMPTY_FEATURES,
                            odl_features.feature_configs)
        # now we do it again, to make sure that it assigns it to default
        # values
        odl_features.deinit()
        self.assertEqual(odl_features.EMPTY_FEATURES,
                         odl_features.feature_configs)

    def test_deinit_resets_to_default_features(self):
        odl_features.deinit()
        self.assertEqual(odl_features.feature_configs,
                         odl_features.EMPTY_FEATURES)



# Copyright (c) 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.db import api as db_api

from networking_odl.common import constants as odl_const
from networking_odl.common import odl_features
from networking_odl.db import db
from networking_odl.qos import qos_driver_v2 as qos_driver
from networking_odl.tests import base
from networking_odl.tests.unit import base_v2


class OpenDaylightQosDriverTestCase(base_v2.OpenDaylightConfigBase):

    def setUp(self):
        self.useFixture(base.OpenDaylightJournalThreadFixture())
        super(OpenDaylightQosDriverTestCase, self).setUp()
        self.qos_driver = qos_driver.OpenDaylightQosDriver({})
        self.addCleanup(odl_features.deinit)

    def test_qos_supported_rules_are_fetched_from_odl_feature(self):
        feature_json = """{"features": {"feature":
                                [{"service-provider-feature":
                                "neutron-extensions:operational-port-status"},
                                {"service-provider-feature":
                                "neutron-extensions:qos-rules",
                                "configuration": {"key": "value"}}]}}"""

        self.cfg.config(odl_features_json=feature_json, group='ml2_odl')
        odl_features.init()

        qos_driver_object = qos_driver.OpenDaylightQosDriver.create()

        self.assertDictEqual(qos_driver_object.supported_rules,
                             {'key': 'value'})

    def test_default_values_for_supported_rules(self):
        self.cfg.config(odl_features='key', group='ml2_odl')
        odl_features.init()

        qos_driver_object = qos_driver.OpenDaylightQosDriver.create()

        self.assertDictEqual(qos_driver_object.supported_rules,
                             qos_driver.DEFAULT_QOS_RULES)

    def _get_mock_context(self, session=None):
        current = {'tenant_id': 'tenant_id'}
        context = mock.Mock(current=current)
        context.session = session
        return context

    def _get_mock_qos_operation_data(self):
        data = {'description': 'qos_policy',
                'rules': [],
                'tenant_id': 'test-tenant',
                'shared': False,
                'id': 'qos-policy1',
                'name': 'policy1'}
        qos_data = mock.Mock()
        to_dict = mock.Mock(return_value=data)
        qos_data.to_dict = to_dict
        return qos_data

    def _call_operation_object(self, operation, object_type):
        qos_data = self._get_mock_qos_operation_data()
        method = getattr(self.qos_driver, '%s_%s' % (operation,
                                                     object_type))

        assert object_type.endswith("precommit")
        with db_api.CONTEXT_WRITER.using(self.db_context):
            context = self._get_mock_context(self.db_context.session)
            method(context, qos_data)

    def _test_qos_policy(self, operation):
        self._call_operation_object(operation=operation,
                                    object_type='policy_precommit')
        qos_data = self._get_mock_qos_operation_data()
        row = db.get_oldest_pending_db_row_with_lock(self.db_context)
        self.assertEqual(operation, row['operation'])
        self.assertEqual(qos_data.to_dict()['id'], row['object_uuid'])

    def test_qos_policy_create(self):
        self._test_qos_policy(odl_const.ODL_CREATE)

    def test_qos_policy_update(self):
        self._test_qos_policy(odl_const.ODL_UPDATE)

    def test_qos_policy_delete(self):
        self._test_qos_policy(odl_const.ODL_DELETE)



# Copyright (c) 2016 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from networking_odl.ml2 import legacy_port_binding
from networking_odl.ml2 import port_binding
from networking_odl.tests import base


class TestPortBindingManager(base.DietTestCase):

    def test_create(self):
        mgr = port_binding.PortBindingManager.create(
            name="legacy-port-binding")
        self.assertEqual("legacy-port-binding", mgr.name)
        self.assertIsInstance(mgr.controller,
                              legacy_port_binding.LegacyPortBindingManager)

    def test_create_with_nonexist_name(self):
        self.assertRaises(AssertionError,
                          port_binding.PortBindingManager.create,
                          name="nonexist-port-binding")

    @mock.patch.object(legacy_port_binding.LegacyPortBindingManager,
                       "bind_port")
    def test_bind_port(self, mock_method):
        port_context = mock.Mock()
        mgr = port_binding.PortBindingManager.create(
            name="legacy-port-binding")
        mgr.controller.bind_port(port_context)
        mock_method.assert_called_once_with(port_context)
